import logging
import sys
import os
import numpy as np
import traceback
file_path = os.path.realpath(__file__)
plugin_dir = os.path.dirname(file_path)
os.chdir(plugin_dir)
#from xframe import dependency_inject_SOFT
#dependency_inject_SOFT()

from xframe.library import mathLibrary as mLib
from xframe.library.mathLibrary import SampleShapeFunctions
from xframe.library.mathLibrary import get_soft_obj
from xframe.library.mathLibrary import PolarIntegrator
from xframe.library.gridLibrary import uniformGrid_func
from xframe.library.gridLibrary import ReGrider
from xframe.library.gridLibrary import NestedArray
from xframe.library.gridLibrary import GridFactory
from xframe.library.pythonLibrary import plugVariableIntoFunction,create_threshold_projection
from xframe.library.pythonLibrary import selectElementOfFunctionOutput_decorator
from xframe.library.pythonLibrary import xprint
from xframe.interfaces import ProjectWorkerInterface

#from .projectLibrary.hankel_transforms import generateHT_SinCos
from .projectLibrary.hankel_transforms import generate_weightDict
from .projectLibrary.fourier_transforms import generate_ft
from .projectLibrary.misk import getAnalysisRecipeFacotry
from .projectLibrary.misk import get_analysis_process_factory
from .projectLibrary.misk import generate_load_from_dict
from .projectLibrary.misk import generate_calc_center
from .projectLibrary.classes import FTGridPair
from .projectLibrary.fxsShrinkWrap import generateSW_SupportMaskRoutine
from .projectLibrary.fxsShrinkWrap import generateSW_multiply_ft_gaussian
from .projectLibrary.fxs_Projections import generate_negative_shift_operator,generate_shift_by_operator
from .projectLibrary.harmonic_transforms import HarmonicTransform
from .projectLibrary.misk import _get_reciprocity_coefficient
from .projectLibrary.resolution_metrics import PRTF_fxs,FQCB_2D,FQCB_3D,FSC_single_fxs,FSC_bit_limit
from .projectLibrary.fxs_invariant_tools import intensity_to_deg2_invariant,harmonic_coeff_to_deg2_invariants
from xframe.externalLibraries import shtns_plugin,soft_plugin
#from analysisLibrary.classes import ReciprocalProjectionData
from xframe import settings
from xframe import database
from xframe import Multiprocessing



##### only for testing, this breaks dependency scheme !!!
from xframe import Multiprocessing

log=logging.getLogger('root')

#class Worker(RecipeInterface):
class ProjectWorker(ProjectWorkerInterface):
    def __init__(self):
        comm_module = Multiprocessing.comm_module        
        self.opt = settings.project
        self.db = database.project
        self.process_factory=get_analysis_process_factory(self.db,[])
        #self.mtip=MTIP(self.process_factory,database,comm_module,opt)
        #self.mp_opt=opt['multi_process']
        #self.mtip_opt=opt
        self.comm_module=comm_module
                
    def post_processing(self,result,ft_grids,r_opt):
        log.info('Start post processing.')
        save = self.db.save
        try:
            reciprocity_coefficient = _get_reciprocity_coefficient(r_opt.fourier_transform)
            result['reciprocity_coefficient'] = reciprocity_coefficient
            result['internal_grid'] = {'real_grid':ft_grids.realGrid[:],
                                       'reciprocal_grid':ft_grids.reciprocalGrid[:]}
            #    if self.mp_opt['type'] != 'routine_construction_and_phasing_and_post_processing':
            #        processed_results=self.results.get('MTIP',{})
            #        processed_results_list=[result_dict for result_dict in processed_results]
            #        #log.info([type(result_dict) for result_dict in processed_results])
            #        #log.info([result_dict.shape for result_dict in processed_results])
            #        errors = []
            #        for result_dict in processed_results_list:
            #            grid_pair = result_dict.pop('grid_pair')
            #            errors.append(result_dict['error_dict']['real_rel_mean_square'][-1])
            #            r_ids=np.argsort(errors)
            #        log.info('error sorted reconstruction = {} \n errors ={}'.format(r_ids,np.array(errors)[r_ids]))
            #        processed_results_dict={str(_id):processed_results_list[_id] for _id in r_ids}
            #        data_dict={'configuration':{'internal_grid':grid_pair,'opt':self.mtip.opt.__dict__},'reconstruction_results':processed_results_dict}
            save('average_results',result)
        except Exception as e:
            traceback.print_exc()
            log.error(e)
    def get_averaged_projection_matrices(self,projection_matrices,average_scaling_factors):
        averaged_matrices = []
        n_matrices = len(projection_matrices[0])
        N_datasets = len(projection_matrices) 
        for i in range(n_matrices):
            temp = projection_matrices[0][i]/average_scaling_factors[0]**2
            for j in range(1,N_datasets):
                temp += projection_matrices[j][i]/average_scaling_factors[j]**2
            averaged_matrices.append(temp/N_datasets)
        return averaged_matrices
            
    def run(self):
        xprint('load reconstructions.')
        reconstructions,masks,r_opt,errors,start_ids,reference_arg,reconstructions_keys,projection_matrices,reconstruction_ids_per_file = getattr(self,self.opt['load_routine'])()        
        xprint('done.\n')
        if len(reconstructions)<2:
            log.error(f'Less than 2 reconstructions where found in the specified reconstruction files:\n {settings.project.reconstruction_files}\n There is nothing to average, abborting calculation.')
            return
        
        if r_opt['GPU'].get('use',False):
            settings.general.n_control_workers = r_opt['GPU'].get("n_gpu_workers",4)
            Multiprocessing.comm_module.restart_control_worker()
        n_reconstructions = self.opt['selection'].get('n_reconstructions','all')
        if not isinstance(n_reconstructions,int) or isinstance(n_reconstructions,bool):
            n_reconstructions = len(reconstructions)
            
        dim = r_opt.dimensions
        if dim == 2:
            result,_locals = self.run_2d(reconstructions,masks,r_opt,errors,start_ids,reference_arg,reconstructions_keys,n_reconstructions,projection_matrices,reconstruction_ids_per_file)
        elif dim == 3:
            result,_locals = self.run_3d(reconstructions,masks,r_opt,errors,start_ids,reference_arg,reconstructions_keys,n_reconstructions,projection_matrices,reconstruction_ids_per_file)
        return result,_locals
            
    def run_2d(self,reconstructions,masks,r_opt,errors,start_ids,reference_arg,reconstructions_keys,n_reconstructions,projection_matrices,reconstruction_ids_per_file):
        opt = self.opt
        xprint('setting up alignment operators.')
        #log.info('reference arg = {}'.format(reference_arg))
        align_opt = {'opt':opt,'r_opt':r_opt}
        alignment = Alignment(self.process_factory,self.db,align_opt)
        fft = self.process_factory.get_operator('fourier_transform')
        ifft = self.process_factory.get_operator('inverse_fourier_transform')
        cht = self.process_factory.get_operator('complex_harmonic_transform')
        icht = self.process_factory.get_operator('complex_inverse_harmonic_transform')               
        xprint('done.\n')
        
        if opt.get('center_reconstructions',False):            
            xprint('centering reconstructions.')
            if opt['multi_process']['use']:
                def mp_center(r_id,**kwargs):
                    alignment = Alignment(self.process_factory,self.db,align_opt)
                    r = reconstructions[r_id]
                    temp = alignment.shift_to_center(*r)                    
                    return [r_id,temp]
                Multiprocessing.MPMode_Queue(assemble_outputs=True)
                outs = Multiprocessing.comm_module.request_mp_evaluation(mp_center,input_arrays=[np.arange(len(reconstructions))],call_with_multiple_arguments=False,split_mode = 'modulus')
                
                #log.info('len outs[0] = {}'.format(len(outs[0])))
                #log.info('len outs = {}'.format(type(outs)))
                for r_id,temp in outs:
                    reconstructions[r_id] = temp[:2]
            else:
                for r_id,r in enumerate(reconstructions):
                    #log.info([d.shape for d in r])
                    temp = alignment.shift_to_center(*r)
                    if opt['use_masks']:
                        shifted_mask_threshold = self.opt['shifted_mask_threshold']
                        shift = mLib.cartesian_to_spherical(-mLib.spherical_to_cartesian(temp[-1]))
                        log.info('shift = {}, - shift = {}'.format(temp[-1], shift))
                        masks[r_id] = alignment.shift_by(masks[r_id],shift)[0]
                        density = temp[0]
                        density[masks[r_id]<shifted_mask_threshold] = 0
                        #temp[0] = density
                    reconstructions[r_id] = temp[:2]
            xprint('done.\n')
            
        scaling_factors = np.ones(len(reconstructions))
        if opt.normalize_reconstructions.use:
            xprint('normalizing reconstructions.')            
            if opt.normalize_reconstructions.mode == 'max':
                for r_id,r in enumerate(reconstructions):
                    xprint(f'Reconstruction {r_id} has real max {np.max(r[0])} and mean {np.mean(r[0])}')
                    scale = np.max(r[0][r[0]>0].real)
                    scaling_factors[r_id]=scale
                    #log.info('{}{}'.format(scale,scale2))
                    reconstructions[r_id] = [r[0]/scale,r[1]/scale]
                    
            elif opt.normalize_reconstructions.mode == 'mean':
                for r_id,r in enumerate(reconstructions):
                    scale = np.mean(r[0][r[0]>0])
                    scaling_factors[r_id]=scale
                    reconstructions[r_id] = [r[0]/scale,r[1]/scale]
            xprint('done.\n')

        average_scaling_factors_per_file = np.ones(len(reconstruction_ids_per_file))
        for file_id,r_ids in enumerate(reconstruction_ids_per_file):
            #log.info('ids per file = {}'.format(r_ids))
            average_scaling_factors_per_file[file_id] = np.mean(scaling_factors[r_ids])
                
        ft_grids = alignment.ft_grids
        rgrid = ft_grids.reciprocalGrid[:]
        integrate = PolarIntegrator(rgrid).integrate
        #log.info('leen reconstructions before reference extraction = {}'.format(len(reconstructions)))
        reference_reconstruction = reconstructions.pop(reference_arg)
        #log.info('leen reconstructions after reference extraction = {}'.format(len(reconstructions)))
        reference_mask = masks.pop(reference_arg)
        d_ref = reference_reconstruction[0].copy()
        ft_d_ref = reference_reconstruction[1].copy()
       
        if opt.get("pointinvert_reference",False):
            xprint('\tpoint inverting reference.')
            #(ift(signal[1].conj()),signal[1].conj())
            ft_d_ref = ft_d_ref.conj()
            d_ref = ifft(fft(d_ref).conj())
            #reference_d = ifft(reference_i.conj())
            #reference_reconstruction = [reference_d,reference_i]
            reference_mask[:]=True #Propper mask handling not implemented for pointinverse
            xprint('done.\n')
        
        aligned_reconstructions = [[d_ref,ft_d_ref]]
        p_inv_metric = []
        ft_d_ref_pinv = ft_d_ref.conj()
        #aligned_reconstructions = [[ifft(ft_d_ref.copy()),ft_d_ref]]
        #log.info('len reconstructions before pinv correction = {}'.format(len(reconstructions)))
        xprint('correct point inversion by comparing to reference.')
        for i,reconstruction in enumerate(reconstructions):
            d,ft_d = reconstruction            
            diff = integrate(np.abs(ft_d.imag-ft_d_ref.imag))
            diff_inv = integrate(np.abs(ft_d.imag-ft_d_ref_pinv.imag))                        
            point_inverted = diff>diff_inv
            #log.info('{} : diff = {} diff inv = {} diff2 = {} diff inv2 = {}'.format(i,diff,diff_inv,diff2,diff_inv2))
            #log.info('diff_reff = {}'.format(integrate(np.abs(ft_d_ref.imag-ft_d_ref_pinv.imag))))            
            if point_inverted:
                new_ft_d = ft_d.copy().conj()
                new_d = ifft(fft(d).conj())
            else:
                new_ft_d = ft_d
                new_d = d

            aligned_reconstructions.append([new_d,new_ft_d])
            p_inv_metric.append([diff,diff_inv,int(point_inverted)])
        xprint('done.\n')
        
        if len(aligned_reconstructions)>=n_reconstructions:
            aligned_reconstructions=aligned_reconstructions[:n_reconstructions]
        log.info('averaging over {} '.format(len(aligned_reconstructions)))

        #density_array = np.array(tuple(r[0] for r in aligned_reconstructions))
        #log.info('density_array.shape = {}'.format(density_array))
        average_density = np.mean(np.array(tuple(r[0] for r in aligned_reconstructions)),axis = 0)
        average_ft_density = np.mean(np.array(tuple(r[1] for r in aligned_reconstructions)),axis = 0)
        aligned_ft_reconstr_ftd = [fft(r[0]) for r in aligned_reconstructions]
        intensity_from_ft_density = np.mean([(r[1]*r[1].conj()).real for r in aligned_reconstructions], axis = 0)
        intensity_from_density= np.mean([(r*r.conj()).real for r in aligned_ft_reconstr_ftd], axis = 0)
        input_reconstructions = [reference_reconstruction,*reconstructions]
        input_masks = [reference_mask,*masks]

        centered_average = alignment.shift_to_center(average_density,average_ft_density)

        xprint('Calculate resolution metrics.')
        resolution_metrics = {}
        if opt.resolution_metrics.get('PRTF',False):
            prtf = PRTF_fxs(fft(average_density),intensity_from_density,averaged_projected_scattering_amplitude=average_ft_density,averaged_projected_intensity=intensity_from_ft_density)
            resolution_metrics['PRTF'] = prtf[0]
            resolution_metrics['PRTF_std'] = prtf[1]
            prtf_d = PRTF_fxs(fft(average_density),intensity_from_density)
            resolution_metrics['PRTF_from_density'] = prtf_d[0]
            resolution_metrics['PRTF_from_density_std'] = prtf_d[1]
            prtf_ftd = PRTF_fxs(average_ft_density,intensity_from_ft_density)
            resolution_metrics['PRTF_from_ft_density'] = prtf_ftd[0]
            resolution_metrics['PRTF_from_ft_density_std'] = prtf_ftd[1]

            prtf_ftI= PRTF_fxs(fft(average_density),intensity_from_ft_density)
            resolution_metrics['PRTF_ftI'] = prtf_ftI[0]
            resolution_metrics['PRTF_ftI_std'] = prtf_ftI[1]
            #log.info('PRTF shape {} PRTF dtype = {}'.format(prtf[0].shape,prtf[0].dtype))
            #log.info('PRTF {}'.format(prtf[0]))
        if opt.resolution_metrics.get('FQCB',False):
            ft_d = fft(average_density)
            b_d_rec=intensity_to_deg2_invariant(ft_d*ft_d.conj(),cht,2)
            b_ftd_rec=intensity_to_deg2_invariant(average_ft_density*average_ft_density.conj(),cht,2)
            pr_mat=np.squeeze(np.array(self.get_averaged_projection_matrices(projection_matrices,average_scaling_factors_per_file)))
            b_target = harmonic_coeff_to_deg2_invariants(2,pr_mat.T)

            #for i in range(len(b_target)):
            #    b_di=b_d_rec[i]
            #    b_ftdi=b_ftd_rec[i]
            #    b_ti=b_target[i]
            #    
            #    b_di[np.abs(b_di)<np.abs(b_di).max()/1e15]=0
            #    b_ftdi[np.abs(b_ftdi)<np.abs(b_ftdi).max()/1e15]=0
            #    b_ti[np.abs(b_ti)<np.abs(b_ti).max()/1e15]=0
                
            fqcb_d = FQCB_2D(b_d_rec,b_target,skip_odd_orders=True)
            fqcb_ftd = FQCB_2D(b_ftd_rec,b_target,skip_odd_orders=True)
            fqcb_rec = FQCB_2D(b_ftd_rec,b_d_rec,skip_odd_orders=True)
            fqcb_d_z = FQCB_2D(b_d_rec,b_target,skip_odd_orders=True,include_zero_order = True)
            fqcb_ftd_z = FQCB_2D(b_ftd_rec,b_target,skip_odd_orders=True,include_zero_order = True)
            resolution_metrics['FQCB_from_density']=fqcb_d[0]
            resolution_metrics['FQCB_from_density_std']=fqcb_d[1]
            resolution_metrics['FQCB_from_ft_density']=fqcb_ftd[0]
            resolution_metrics['FQCB_from_ft_density_std']=fqcb_ftd[1]
            resolution_metrics['FQCB_projected_vs_unprojected']=fqcb_rec[0]
            resolution_metrics['FQCB_projected_vs_unprojected_std']=fqcb_rec[1]
            
            resolution_metrics['FQCB_from_density_with_zero_order']=fqcb_d_z[0]
            resolution_metrics['FQCB_from_density_with_zero_order_std']=fqcb_d_z[1]
            resolution_metrics['FQCB_from_ft_density_with_zero_order']=fqcb_ftd_z[0]
            resolution_metrics['FQCB_from_ft_density_with_zero_order_std']=fqcb_ftd_z[1]
            
            resolution_metrics['FQCB_bl_from_density']=b_d_rec
            resolution_metrics['FQCB_bl_from_ft_density']=b_ftd_rec
            resolution_metrics['FQCB_bl_target']=b_target
            
        if opt.resolution_metrics.get('pseudo_FSC',False):
            pseudo_fsc = FSC_single_fxs(fft(average_density),average_ft_density)
            resolution_metrics['pseudo_FSC']=pseudo_fsc
            fsc_limit = FSC_bit_limit(0.5,rgrid)
            resolution_metrics['FSC_0.5bit_limit']=fsc_limit
            #input_b = 
            
        xprint('done.\n')

        xprint('Averaging completed saving results.')
        result = {
            'average':
            {
                'real_density':average_density,'normalized_real_density':self.normalize_density(average_density),'reciprocal_density':average_ft_density,'intensity_from_densities':intensity_from_density,'intensity_from_ft_densities':intensity_from_ft_density
            },
            'resolution_metrics':resolution_metrics,
            'centered_average':
            {
                'real_density':centered_average[0],'normalized_real_density':self.normalize_density(centered_average[0]),'reciprocal_density':centered_average[1]
            },
            "aligned":
            {
                #**{
                #'0':{'real_density':reference_reconstruction[0],'reciprocal_density':reference_reconstruction[0]},
                #'1':{'real_density':d,'reciprocal_density':a},
                #},
                
                **{
                    str(i):{'real_density':r[0],'reciprocal_density':r[1]} for i,r in enumerate(aligned_reconstructions)
                },
                
                #**{
                #str(i+2):{'real_density':reconstructions[i][0],'reciprocal_density':reconstructions[i][1]} for i in range(len(reconstructions))
                #}
            },
            'input':
            {
                str(i):{'real_density':input_reconstructions[i][0],
                        'reciprocal_density':input_reconstructions[i][1],
                        'support_mask':input_masks[i]} for i in range(len(input_reconstructions))
            },
            'input_meta':{
                'projection_matrices':projection_matrices,
                'reconstruction_keys':reconstructions_keys,
                'scaling_factors': np.array(scaling_factors),
                'average_scaling_factors_per_file':np.array(average_scaling_factors_per_file),
                'grids':ft_grids
            },
            'inversion_metric':
            {
                str(i+1): np.array(m) for i,m in enumerate(p_inv_metric)         
            }
        }
        self.post_processing(result,ft_grids,r_opt)
        result = 0
        return result,locals()
    def run_3d(self,reconstructions,masks,r_opt,errors,start_ids,reference_arg,reconstructions_keys,n_reconstructions,projection_matrices,reconstruction_ids_per_file):
        opt = self.opt        
        #log.info('reference arg = {}'.format(reference_arg))
        xprint('setting up alignment operators.')
        align_opt = {'opt':opt,'r_opt':r_opt}
        alignment = Alignment(self.process_factory,self.db,align_opt)
        ft_grids = alignment.ft_grids
        rgrid = ft_grids.reciprocalGrid[:]
        fft = self.process_factory.get_operator('fourier_transform')
        ifft = self.process_factory.get_operator('inverse_fourier_transform')
        xprint('done.\n')
        
        #xprint(f'sums {[np.sum(r[0]) for r in reconstructions]}')
        if opt.get('center_reconstructions',False):
            xprint('centering reconstructions.')
            if opt['multi_process']['use']:
                n_processes = opt['multi_process']['n_processes']
                log.info(f'n_processes centering {n_processes}')
                def mp_center(r_ids,**kwargs):
                    #temeporary fix to define alignment here its align_to function has problems with forking.
                    #the shift center part is likely not affected but just to be sure.
                    process_factory = get_analysis_process_factory(self.db,[])
                    alignment = Alignment(process_factory,self.db,align_opt)
                    results = []
                    for r_id in r_ids:
                        r = reconstructions[r_id]
                        temp = alignment.shift_to_center(*r)
                        results.append([r_id,temp])
                    return results
                Multiprocessing.MPMode_Queue(assemble_outputs=True)
                outs = Multiprocessing.comm_module.request_mp_evaluation(mp_center,input_arrays=[np.arange(len(reconstructions))],call_with_multiple_arguments=True,split_mode = 'modulus',n_processes = n_processes)
                
                #log.info('len outs[0] = {}'.format(len(outs[0])))
                #log.info('len outs = {}'.format(type(outs)))
                #for o in outs:
                #    log.info(f'len part = {len(o)}')
                #log.info('len outs[0] = {}'.format(len(outs[0])))
                #log.info('len outs = {}'.format(len(outs)))
                for temp in outs:
                    for r_id,result in temp:
                        reconstructions[r_id] = result[:2]
                #for r_id,temp in outs:
                #    reconstructions[r_id] = temp[:2]
            else:
                for r_id,r in enumerate(reconstructions):
                    #log.info([d.shape for d in r])
                    temp = alignment.shift_to_center(*r) 
                    if opt['use_masks']:
                        shifted_mask_threshold = self.opt['shifted_mask_threshold']
                        shift = mLib.cartesian_to_spherical(-mLib.spherical_to_cartesian(temp[-1]))
                        log.info('shift = {}, - shift = {}'.format(temp[-1], shift))
                        masks[r_id] = alignment.shift_by(masks[r_id],shift)[0]
                        density = temp[0]
                        density[masks[r_id]<shifted_mask_threshold] = 0
                        #temp[0] = density
                    reconstructions[r_id] = temp[:2]
            xprint('done.\n')
            #xprint(f'sums {[np.sum(r[0]) for r in reconstructions]}')
       
       
            
        #log.info('max mask unshifted = {}'.format(reference_mask.max()))
        #n_reconstructions= opt['selection']['n_reconstructions']
        #reconstructions = [ reconstructions[arg] for arg in np.argsort(errors)[:n_reconstructions-1]]                
        #log.info('mask shape = {}'.format(ref_mask.shape))
        #ref_mask = np.array(reference_mask,dtype=np.complex128)
        #log.info('mask shape = {}'.format(ref_mask.shape))
        #ref_r_mask = fft(ref_mask)
        #log.info('alive')

        scaling_factors = np.ones(len(reconstructions))
        if opt.normalize_reconstructions.use:
            xprint('normalizing reconstructions.')
            if opt.normalize_reconstructions.mode == 'max':                
                for r_id,r in enumerate(reconstructions):
                    if np.max(r[0]).real<=0:                        
                        xprint(f'Reconstruction {r_id} has zero max! continue.')
                        continue
                    scale = np.max(r[0][r[0]>0].real)
                    #xprint(f'Reconstruction {r_id} has max {scale} ref is {reference_arg}')
                    scaling_factors[r_id]=scale
                    #log.info('{}{}'.format(scale,scale2))
                    reconstructions[r_id] = [r[0]/scale,r[1]/scale]
            
            elif opt.normalize_reconstructions.mode == 'mean':
                for r_id,r in enumerate(reconstructions):
                    scale = np.mean(r[0][r[0]>0])
                    scaling_factors[r_id]=scale
                    reconstructions[r_id] = [r[0]/scale,r[1]/scale]
            xprint('done.\n')        
        average_scaling_factors_per_file = np.ones(len(reconstruction_ids_per_file))
        for file_id,r_ids in enumerate(reconstruction_ids_per_file):
            #log.info('ids per file = {}'.format(r_ids))
            average_scaling_factors_per_file[file_id] = np.mean(scaling_factors[r_ids])

        reference_reconstruction = reconstructions.pop(reference_arg)
        reference_mask = masks.pop(reference_arg)
        #log.info("opt pinv =  {}".format(opt.get("pointinvert_reference",False)))
        #log.info("opt =  {}".format(opt))
        if opt.get("pointinvert_reference",False):
            xprint('\tpoint inverting reference.')
            #(ift(signal[1].conj()),signal[1].conj())
            reference_i = reference_reconstruction[1].conj()
            reference_d = ifft(reference_i)
            reference_reconstruction = [reference_d,reference_i]
            reference_mask[:]=True #Propper mask handling not implemented for pointinverse
            xprint('done.\n')
        #       
        #        n_reconstructions = len(reconstructions)
        
        #reference_reconstruction = alignment.shift_by(reconstructions[0][0],np.array((100,np.pi/2,0)))
        #reference_reconstruction = alignment.rotate_by(r2[0],np.array([0.5*np.pi,0.7*np.pi,1.5*np.pi])) #alignment.shift_by(r2[0],np.array((100,np.pi/2,0)))
        #print('ref = {} sig = {}'.format(2,2))
        #reference_reconstruction = r2 #alignment.shift_by(r2[0],np.array((100,np.pi/2,0)))
        
        xprint('rotationaly aligning reconstructions to reference.')
        outs = []
        valid_errors = []
        densities = [reference_reconstruction]
        rotation_metrics = []
        rotation_angles = []
        valid_alignments= [reference_reconstruction]
        valid_alignment_ids = [0]
        alignment_error_limit = settings.project.alignment_error_limit        
        if opt['multi_process']['use']:
            n_processes = opt['multi_process']['n_processes']
            log.info(f'n_processes align {n_processes}')
            def mp_align_pattern(_ids,**kwargs):
                results = []
                #temeporary fix: We create a new alignment object here because its apply_to function has problems with forking.
                alignment = Alignment(self.process_factory,self.db,align_opt)
                for _id in _ids:
                    result = [_id,alignment.apply_to(reference_reconstruction[0].copy(),reconstructions[_id])]
                    results.append(result)
                del(alignment)
                return results
            Multiprocessing.MPMode_Queue(assemble_outputs=True)
            outs = Multiprocessing.comm_module.request_mp_evaluation(mp_align_pattern,input_arrays=[np.arange(len(reconstructions))],call_with_multiple_arguments=True,split_mode='modulus',n_processes = n_processes)
            errors = []
            for temp in outs:
                for out in temp:
                    #log.info(out.keys())
                    out = out[1]
                    densities.append(out['densities'])
                    rotation_metrics.append(out['rotation_metrics'])
                    rotation_angles.append(out['rotation_angles'])
                    #log.info('error = {}'.format(out['errors'][-1]))
                    if out['errors'][-1] < alignment_error_limit:
                        valid_errors.append(out['errors'][-1])
                        valid_alignments.append(out['densities'])
                        valid_alignment_ids.append(r_id)
                    errors.append(out['errors'][-1])
        else:
            for r_id,reconstruction in enumerate(reconstructions):
                out = alignment.apply_to(reference_reconstruction[0].copy(),reconstruction)
                outs.append(out)
                densities.append(out['densities'])
                rotation_metrics.append(out['rotation_metrics'])
                rotation_angles.append(out['rotation_angles'])
                #log.info('error = {}'.format(out['errors'][-1]))
                if out['errors'][-1] < alignment_error_limit:
                    valid_errors.append(out['errors'][-1])
                    valid_alignments.append(out['densities'])
                    valid_alignment_ids.append(r_id)
            errors = [out['errors'][-1] for out in outs]

        xprint('done.\n')        
        sorted_errors = np.argsort(valid_errors)
        aligned_reconstr = [valid_alignments[i] for i in sorted_errors]
        if len(aligned_reconstr)>=n_reconstructions:
            aligned_reconstr=aligned_reconstr[:n_reconstructions]
        else:
            log.info('Only {} reconstructions of the wanted {} satisfy error constraint. Averaging over viewer reconstructions than targeted.'.format(len(valid_errors),n_reconstructions))

        xprint('Averaging over {} aligned patterns with error smaller than {}!'.format(len(aligned_reconstr),alignment_error_limit))
        xprint('allignment errors are = {}\n'.format(np.sort(errors)))
            
        #average = [np.mean(d,axis = 0) for d in zip(*aligned_reconstr)]
        average = np.squeeze(np.mean(aligned_reconstr,axis = 0))
        aligned_ft_reconstr_ftd = [fft(r[0]) for r in aligned_reconstr]
        intensity_from_ft_density = np.mean([(r[1]*r[1].conj()).real for r in aligned_reconstr], axis = 0)
        intensity_from_density= np.mean([(r*r.conj()).real for r in aligned_ft_reconstr_ftd], axis = 0)
        input_reconstructions = [reference_reconstruction,*reconstructions]
        input_masks = [reference_mask,*masks]
        #log.info("{}, {}".format(len(input_reconstructions),len(input_masks)))

        centered_average = alignment.shift_to_center(*average)
        average_normalization_min = opt.get('average_normalization_min',False)

        xprint('calculating resolution metrics.')
        resolution_metrics = {}
        resolution_metrics = {}
        if opt.resolution_metrics.get('PRTF',False):
            prtf = PRTF_fxs(fft(average[0]),intensity_from_density,averaged_projected_scattering_amplitude=average[1],averaged_projected_intensity=intensity_from_ft_density)
            resolution_metrics['PRTF'] = prtf[0]
            resolution_metrics['PRTF_std'] = prtf[1]
            prtf_d = PRTF_fxs(fft(average[0]),intensity_from_density)
            resolution_metrics['PRTF_from_density'] = prtf_d[0]
            resolution_metrics['PRTF_from_density_std'] = prtf_d[1]
            prtf_ftd = PRTF_fxs(average[1],intensity_from_ft_density)
            resolution_metrics['PRTF_from_ft_density'] = prtf_ftd[0]
            resolution_metrics['PRTF_from_ft_density_std'] = prtf_ftd[1]

            prtf_ftI= PRTF_fxs(fft(average[0]),intensity_from_ft_density)
            resolution_metrics['PRTF_ftI'] = prtf_ftI[0]
            resolution_metrics['PRTF_ftI_std'] = prtf_ftI[1]
            #log.info('PRTF shape {} PRTF dtype = {}'.format(prtf[0].shape,prtf[0].dtype))
            #log.info('PRTF {}'.format(prtf[0]))
        if opt.resolution_metrics.get('FQCB',False):
            pass            
        if opt.resolution_metrics.get('pseudo_FSC',False):
            pseudo_fsc = FSC_single_fxs(fft(average[0]),average[1])
            resolution_metrics['pseudo_FSC']=pseudo_fsc
            fsc_limit = FSC_bit_limit(0.5,rgrid)
            resolution_metrics['FSC_0.5bit_limit']=fsc_limit
        xprint('done.\n')

        xprint('Saving results.\n')
        result = {
            'average':
            {
                'real_density':average[0],'normalized_real_density':self.normalize_density(average[0],d_min=average_normalization_min),'reciprocal_density':average[1],'intensity_from_densities':intensity_from_density,'intensity_from_ft_densities':intensity_from_ft_density
            },
            'resolution_metrics':resolution_metrics,
            'centered_average':
            {
                'real_density':centered_average[0],'normalized_real_density':self.normalize_density(centered_average[0],d_min=average_normalization_min),'reciprocal_density':centered_average[1]
            },
            "average_ids":valid_alignment_ids,
            "aligned":
            {
                #**{
                #'0':{'real_density':reference_reconstruction[0],'reciprocal_density':reference_reconstruction[0]},
                #'1':{'real_density':d,'reciprocal_density':a},
                #},
                
                **{
                    str(i):{'real_density':aligned_reconstr[i][0],'reciprocal_density':aligned_reconstr[i][1]} for i in range(len(aligned_reconstr))
                },
                
                #**{
                #str(i+2):{'real_density':reconstructions[i][0],'reciprocal_density':reconstructions[i][1]} for i in range(len(reconstructions))
                #}
            },
            'input':
            {
                str(i):{'real_density':input_reconstructions[i][0],'reciprocal_density':input_reconstructions[i][1],'support_mask':input_masks[i]} for i in range(len(input_reconstructions))
            },
            'input_meta':{
                'projection_matrices':projection_matrices,
                'reconstruction_keys':reconstructions_keys,
                'scaling_factors': np.array(scaling_factors),
                'average_scaling_factors_per_file':np.array(average_scaling_factors_per_file),
                'grids':ft_grids
            },
            'rotation_metric':
            {
                str(i+1): rotation_metrics[i] for i in range(len(rotation_metrics))         
            },
            'rotation_angles':
            {
                str(i+1): rotation_angles[i] for i in range(len(rotation_angles))         
            },
            'so3_grid': alignment.soft_grid,
        }
        self.post_processing(result,alignment.ft_grids,r_opt)
        result = 0
        return result,locals()

    
    def load_reconstructions(self):
        opt = self.opt
        db = self.db        
        error_metric = opt['selection']['error_metric']
        error_limit = opt['selection']['error_limit']        
        reconstructions = []
        reconstruction_ids_per_file = [[] for i in range(len(opt['reconstruction_files']))]
        reconstruction_keys = {}
        reconstruction_keys2 = {}
        support_masks = []
        errors = []
        file_start_ids = []
        projection_matrices=[]
        start_id = 0
        r_opt = False
        
        file_paths = [0]*len(opt['reconstruction_files'])
        for num,file_path in enumerate(opt['reconstruction_files']):
            file_path=file_path.format(today=db.get_time_string())
            file_paths[num]=file_path
        
        for num,file_path in enumerate(file_paths):
            xprint('\tfile {} of {}'.format(num+1,len(opt['reconstruction_files'])))
            with db.load('reconstructions',path_modifiers={'path':file_path},as_h5_object = True) as h5_file:
                file_start_ids.append(start_id)
                reconstruction_keys[num] = {}
                start_id += len(h5_file['reconstruction_results'])
                max_order = len(h5_file['projection_matrices'])
                pr_group = h5_file['projection_matrices']
                pr_mat = [pr_group[str(i)][:] for i in range(max_order)]
                projection_matrices.append(pr_mat)
                #log.info(start_id)
                for key,value in h5_file['reconstruction_results'].items():
                    error = value['error_dict'][error_metric][-1]
                    if error < error_limit:
                        reconstr = (value['last_real_density'][:],value['last_reciprocal_density'][:])
                        if self.valid_maximal_density(reconstr[0].real.max()):
                            errors.append(error)
                            reconstruction_keys[num][key] = len(reconstructions)
                            pos = len(reconstructions)
                            reconstruction_keys2[pos] = (num,key)
                            reconstruction_ids_per_file[num].append(pos)
                            reconstructions.append(reconstr)                            
                            support_masks.append(np.array(value['support_mask'][:],dtype=float))
                if num == 0:                    
                    r_opt = db.load('reconstruction_settings',path_modifiers= {'path':os.path.dirname(file_path)})
                    r_opt['internal_grid'] = db.get_db('file://_.h5').recursively_load_dict_from_group(h5_file,'/configuration/')['internal_grid']
        errors = np.array(errors)

        #sort reconstructions by error_metric
                                
                
        #support_masks = [ support_masks[rid] for rid in min_args]
        #errors = errors[min_args]
        
        xprint("\t{} loaded reconstruction with error metrics smaller than {}, error values are: \n\t{}".format(len(reconstructions),error_limit,errors))
        g_dict = r_opt['internal_grid']
        #log.info(g_dict.real_grid)
        ft_grids = FTGridPair(g_dict['real_grid'],g_dict['reciprocal_grid'])
        r_opt['internal_grid'] = ft_grids
        reference_arg = self.get_reference_arg(errors,reconstruction_keys)
        #log.info('reference reconstruction={}'.format(reconstruction_keys2[reference_arg]))
        reference_file_path = db.get_path('reconstructions',path_modifiers={'path':file_paths[reconstruction_keys2[reference_arg][0]]})
        xprint("\tselected reference: reconstruction {} in {}".format(reconstruction_keys2[reference_arg][1],reference_file_path))
        return reconstructions,support_masks,r_opt,errors,file_start_ids,reference_arg,reconstruction_keys2,projection_matrices,reconstruction_ids_per_file

    def extract_aligned_reconstructions(self,alignment_results):
        aligned_reconstructions = []
        for align_dict in alignment_results:
            aligned_reconstructions.append(align_dict['reconstruction'])
        return aligned_reconstructions

    def get_reference_arg(self,errors,lookup_dict):
        selection_method = self.opt['selection']['method']
        if selection_method == 'least_error':
            reconstruction_arg = np.argmin(errors)
        elif selection_method == 'manual':
            m_spec = self.opt['selection']['manual_specifier']
            reconstruction_arg = lookup_dict[m_spec[0]][str(m_spec[1])]
        return reconstruction_arg

    def valid_maximal_density(self,max_density):
        lower_bound,upper_bound = settings.project.selection.get('max_density_range',[None,None])   
        is_valid = True
        if isinstance(lower_bound,(int,float)):
            if max_density < lower_bound:
                is_valid=False
        if isinstance(upper_bound,(int,float)):
            if max_density > upper_bound:
                is_valid=False
        return is_valid

    def normalize_density(self,d,d_min=False):
        if isinstance(d_min,bool):
            d_min = d.real.min()
        #log.info('dmin = {}'.format(d_min ))
        d_max = np.max(d.real)
        d_normalized = (d-d_min)/(d_max-d_min)
        return d_normalized
    
class Alignment():
    def __init__(self,process_factory,db,opt):
        self.opt = opt['opt'] # alignment options
        self.r_opt = opt['r_opt'] # options of the reconstructions
        if self.r_opt['GPU'].get('use',False):
            if settings.general.n_control_workers ==0:
                settings.general.n_control_workers = self.r_opt['GPU'].get("n_gpu_workers",4)
                Multiprocessing.comm_module.restart_control_worker()
        self.ft_grids = self.r_opt.internal_grid
        self.dimension = self.ft_grids.realGrid[:].shape[-1]
        self.db = db
        self.process_factory = process_factory
        self._reference_reconstruction = 'not set'
        self.results = {}
        # Load the library that takes carre of rotations
        if self.dimension ==3:
            self.soft = get_soft_obj(self.r_opt['grid']['max_order']+1)
            self.soft_grid = self.soft.make_SO3_grid()
        self.sht = 'not set' #spherical harmonic transform object will be set by generate_operators
        # The next function call generates all necessary operators and adds them
        # to the process_factory.
        self.generate_operators()
        self.shift_to_center = self.assemble_shift_to_center().run
        self.shift_by = self.assemble_shift_by()
        if self.dimension ==3:
            self.rotate_by = self.assemble_rotate_by()
            self.align = self.assemble_align()
            self.apply_to = self.generate_alignment_loop()
        

    @property
    def reference_reconstruction(self):
        return self._reference_reconstruction
    @reference_reconstruction.setter
    def reference_reconstruction(self,value):
        assert (len(value)==2 and isinstance(value[0],np.ndarray) ),'Reference reconstruction needs to be a tuple of length 2 containing np.ndarrays for real_density and reciprocal_density but {} wsr given'.format(value)
        self._reference_reconstruction = value
        self._update_alignment()

    def _update_alignment(self):        
        self.align = self.assemble_align().run
        self.apply_to = self.generate_alignment_loop()
        

    def generate_operators(self):
        r_opt = self.r_opt
        if self.dimension ==3:
            self.harm_lm_split_ids, self.harm_ml_split_ids,self.sht = self.assemble_transform_op(r_opt)
        elif self.dimension ==2 :
            self.sht = self.assemble_transform_op(r_opt)
        find_center = generate_calc_center(self.ft_grids.realGrid)
        shift = generate_shift_by_operator(self.ft_grids.reciprocalGrid)
        negative_shift = generate_shift_by_operator(self.ft_grids.reciprocalGrid,opposite_direction = True)
        self.process_factory.addOperators({'find_center':find_center,'shift':shift,'negative_shift':negative_shift})
        if self.dimension ==3:
            rotate = self.generate_rotate()
            self.process_factory.addOperators({'rotate':rotate})


    def generate_fourier_transforms(self,grid_pair,ft_opt,harm_trf):
        opt = settings.project
        db = database.project
        dimensions = grid_pair.realGrid[:].shape[-1]
        ft_type = ft_opt['type']
        max_order=self.r_opt['grid']['max_order']
        harmonic_orders = np.arange(max_order+1)
        n_orders=len(harmonic_orders)
        reciprocity_coefficient = _get_reciprocity_coefficient(ft_opt)
        rs=grid_pair.realGrid[:].__getitem__((slice(None),)+(0,)*self.dimension) # expression is just grid[:,0,0,0] in 3D case ans grid[:,0,0] in 2D
        r_max = np.max(rs)
        n_radial_points = len(rs)
    
        name_postfix='N'+str(n_radial_points)+'mO'+str(max_order)+'nO'+str(n_orders)+'rc'+str(reciprocity_coefficient)
        try:
            weights_dict = db.load('ft_weights',path_modifiers={'postfix':name_postfix,'type':ft_type+'_'+str(dimensions)+'D'})
        except FileNotFoundError as e:
            weights_dict = generate_weightDict(max_order, n_radial_points,reciprocity_coefficient=reciprocity_coefficient,dimensions=dimensions,mode=ft_type)
        
        
        use_gpu = self.r_opt['GPU']['use']
        
        fourierTransform,inverseFourierTransform=generate_ft(r_max,weights_dict,harm_trf,dimensions,pos_orders=harmonic_orders,reciprocity_coefficient=reciprocity_coefficient,use_gpu = use_gpu,mode = ft_type)
        return fourierTransform,inverseFourierTransform

    def generate_fourier_transforms_old(self, grid_pair,ft_opt , harm_trf):                    
        db = database.project
        dimensions = grid_pair.realGrid[:].shape[-1]
        ft_type = ft_opt.type
        data_type = ft_opt.data_type
        max_order=self.r_opt['grid']['max_order']
        harmonic_orders = np.arange(max_order+1)
        n_orders=len(harmonic_orders)

        pi_in_q = ft_opt.get('pi_in_q',None)
        reciprocity_coefficient = _get_reciprocity_coefficient(ft_opt)

        rs=grid_pair.realGrid[:].__getitem__((slice(None),)+(0,)*self.dimension) # expression is just grid[:,0,0,0] in 3D case ans grid[:,0,0] in 2D
        r_max = np.max(rs)
        n_radial_points = len(rs)
            
        use_gpu = self.r_opt['GPU']['use']
        name_postfix='N'+str(n_radial_points)+'mO'+str(max_order)+'nO'+str(n_orders)+'rc'+str(reciprocity_coefficient)
        try:
            weights_dict = db.load('ft_weights',path_modifiers={'postfix':name_postfix,'type':ft_type+'_'+str(dimensions)+'D'})
        except FileNotFoundError as e:
            if ft_opt.allow_weight_calculation:
                weights_dict = generate_weightDict(max_order, n_radial_points,reciprocity_coefficient=reciprocity_coefficient,dimensions=dimensions,mode=ft_type)
                if ft_opt.allow_weight_saving:
                    db.save('ft_weights',weights_dict,path_modifiers={'postfix':name_postfix,'type':ft_type+'_'+str(dimensions)+'D'})

        
        fourierTransform,inverseFourierTransform=generate_ft(r_max,weights_dict,harm_trf,dimensions,pos_orders=harmonic_orders,reciprocity_coefficient=reciprocity_coefficient,use_gpu = use_gpu,mode = ft_type)

        #fourierTransform,inverseFourierTransform=generate_ft(r_max,weights_dict,harm_trf,dimensions,reciprocity_coefficient,use_gpu = use_gpu,mode = ft_type)
        
        return fourierTransform,inverseFourierTransform

    def assemble_transform_op(self,r_opt):
        ht_opt={
            'dimensions':self.dimension,
            **r_opt['grid']
        }
        # Create harmonic transforms (needs to happen before grid selection since harmonic transform plugin can choose the angular part of the grid.)
        #log.info('create harmonic transforms.')
        cht=HarmonicTransform('complex',ht_opt)
        #ht_data_type = opt['harmonic_transform']['data_type']
        if self.dimension == 2:
            ht =  HarmonicTransform('real',ht_opt)
        elif self.dimension == 3:
            ht=cht
        cht_forward,cht_inverse = cht.forward, cht.inverse
        ht_forward,ht_inverse = ht.forward, ht.inverse
        #log.info("ht grid param = {}".format(cht.grid_param))
        # fourier transforms
        ft_forward,ft_inverse=self.generate_fourier_transforms(self.ft_grids,r_opt['fourier_transform'],cht)

        transform_op={'fourier_transform':ft_forward,'inverse_fourier_transform':ft_inverse,'harmonic_transform':ht_forward,'inverse_harmonic_transform':ht_inverse,'complex_harmonic_transform':cht_forward,'complex_inverse_harmonic_transform':cht_inverse}
        # fourier transforms on SO(3)
        if self.dimension == 3:
            l_max= np.max(r_opt['grid']['max_order'])
            soft = mLib.get_soft_obj(l_max)
            harmonic_transpose_lm_to_ml, harmonic_transpose_ml_to_lm = self.generate_harmonic_transposes(cht)
            transform_op.update({'soft':soft.forward_cmplx,'inverse_soft':soft.inverse_cmplx,'lm_to_ml':harmonic_transpose_lm_to_ml,'ml_to_lm':harmonic_transpose_ml_to_lm})

        self.process_factory.addOperators(transform_op)
        if self.dimension == 3:
            return cht.l_split_indices,cht.m_split_indices,cht
        elif self.dimension ==2:
            return cht

    def assemble_align_parts(self):         
        find_shift = self.generate_find_shift()
        find_rotation = self.generate_find_rotation()
        save_shift = self.generate_save_shift()
        parts = {'find_shift':find_shift,'find_rotation':find_rotation,'save_shift':save_shift}        
        self.process_factory.addOperators(parts)
        
    def generate_harmonic_transposes(self,cht):
        m_indices_concat = np.concatenate(cht.m_indices)
        m_indices = cht.m_indices
        # also turns a single 1D numpy array into a tuple of 1D arrays
        def lm_to_ml(l_coeff):
            #log.info('l_coeff shape = {}'.format(l_coeff.shape))
            m_coeff = [l_coeff[:,index] for index in m_indices]
            return m_coeff
        # also turns tuple of 1Darrays into a single 1D numpy array
        def ml_to_lm(m_coeff):
            l_coeff = np.zeros((len(m_coeff[0]),cht.n_coeff),np.complex128)
            l_coeff[:,m_indices_concat] = np.concatenate(m_coeff,axis = 1)
            return l_coeff
        return lm_to_ml,ml_to_lm

    
    def generate_find_shift(self):
        ref_amplitude = self.reference_reconstruction[1]
        ift = self.process_factory.get_operator('inverse_fourier_transform')
        real_grid = self.ft_grids.realGrid
        def find_shift(ref_amplitude,amplitude):
            cc = ift(amplitude*ref_amplitude.conj())
            argmax = np.unravel_index(np.argmax(cc),cc.shape)
            shift = real_grid[argmax]
            #log.info('calculated_shift = {}'.format(shift))
            return shift
        return find_shift
    
    def generate_save_shift(self):
        def save_shift(shift):
            shift_list = self.results.get('shifts',[])
            shift_list.append(shift)
            self.results['shifts'] = shift_list
        return save_shift
            
    
    def generate_find_rotation(self):
        bw = self.soft.bw
        C_shape = (2*bw,)*3
        pi = np.pi
        calc_mean_C = self.soft.calc_mean_C
        lm_split_ids = self.harm_lm_split_ids
        angle_grid = self.soft_grid
        r_limit_ids = self.opt['find_rotation'].get('r_limit_ids', [0,self.ft_grids.realGrid.shape[0]])
        self.beta = 0
        def find_rotation(ref_lm_coeff,lm_coeff):            
            ref_lm_coeff = np.concatenate(ref_lm_coeff,axis = 1).copy()
            lm_coeff = np.concatenate(lm_coeff,axis = 1).copy()
            #print('limits = {}'.format(r_limit_ids))
            mean_C = calc_mean_C(lm_coeff,ref_lm_coeff,r_limit_ids,lm_split_ids).real
            self.results['rotation_metrics'].append(mean_C)
            #log.info('mean C shape = {},max = {}, min = {}'.format(mean_C.shape,np.max(mean_C),np.min(mean_C)))
            argmax = np.unravel_index(np.argmax(mean_C),C_shape)
            euler_angles = angle_grid[argmax[1],argmax[0],argmax[2]]
            euler_angles[0] = 2*pi - euler_angles[0]
            euler_angles[2] = 2*pi - euler_angles[2]
            self.results['rotation_angles'].append(euler_angles)
            #log.info('mean_C max ids = {},max_C = {}, min_C = {},dtype = {},mean = {},var = {}'.format(argmax,np.max(mean_C),np.min(mean_C),mean_C.dtype,np.mean(mean_C),np.median(mean_C)))
            #log.info('calculated_rotation angle = {}'.format(euler_angles/np.pi))
            return euler_angles
        return find_rotation    
    def generate_rotate(self):
        rotate_coeff = self.soft.rotate_coeff
        lm_split_ids = self.harm_lm_split_ids
        l_indices = self.sht.l_indices
        def rotate(lm_coeff,euler_angles):
            #log.info('euler_angles = {}'.format(euler_angles))
            lm_coeff = np.concatenate(lm_coeff,axis = 1)
            #log.info(lm_coeff.shape)
            rotated_coeff = rotate_coeff(lm_coeff,lm_split_ids,euler_angles)
            rotated_coeff = np.split(rotated_coeff,lm_split_ids,axis = 1)
            return rotated_coeff
        return rotate

    def assemble_align(self):
        self.assemble_align_parts()        
        # signal : the density to be aligned to the reference
        # reference : the reference density, reference_intensity
        rotate_signal_sketch =[
            ['complex_harmonic_transform','complex_harmonic_transform','complex_harmonic_transform'],
            [np.array([0,1,1,2],dtype=np.int64),['find_rotation','id','id']],
            [np.array([1,0,2,0],dtype=np.int64),['rotate','rotate']],
            ['complex_inverse_harmonic_transform','complex_inverse_harmonic_transform'],
        ]
        
        shift_reference_sketch = [
            [np.array([0,1],dtype=np.int64),['fourier_transform','fourier_transform']],
            [np.array([0,0,1],dtype=np.int64),['id','find_shift']],
            [np.array([0,1,1],dtype=np.int64),['shift','save_shift']],
            [np.array([0],dtype=np.int64),['inverse_fourier_transform']],
        ]
        rotate_signal = self.process_factory.buildProcessFromSketch(rotate_signal_sketch)
        shift_reference = self.process_factory.buildProcessFromSketch(shift_reference_sketch)
        self.process_factory.addOperators({'rotate_signal':rotate_signal,'shift_reference':shift_reference})
        
        align_sketch = [
            [np.array([0,1,2],dtype=np.int64),['rotate_signal']],
            #[np.array([0,1,1],dtype=np.int64),['shift_reference','id']],            
            # calc fourier transforms
            #[np.array([0,0,1,1],dtype=np.int64),['id','fourier_transform','id','fourier_transform']],
        ]
        
        align = self.process_factory.buildProcessFromSketch(align_sketch).run
        return align
    
    def assemble_align_old(self):
        self.assemble_align_parts()
        align_sketch = [
            'id',
            ['id','find_shift'],
            'shift',
            'inverse_fourier_transform',
            'complex_harmonic_transform',
            ['id','find_rotation'],
            'rotate',
            'complex_inverse_harmonic_transform',
            ['id','fourier_transform']                                  
        ]
        align_sketch = [
            # rotational alignment step
            ['complex_harmonic_transform','complex_harmonic_transform'],
            [np.array([0,0,1,1],dtype=np.int64),['id','find_rotation','id']],
            [np.array([0,2,1],dtype=np.int64),['id','rotate']],
            ['complex_inverse_harmonic_transform','complex_inverse_harmonic_transform'],
            [np.array([0,0,1,1],dtype=np.int64),['id','fourier_transform','id','fourier_transform']]
        ]
        align = self.process_factory.buildProcessFromSketch(align_sketch).run
        return align
    def assemble_shift_to_center(self):
        #start with reciprocal_density,real_density
        shift_to_center_sketch_old = [
            ['find_center','id'],
            [np.array([1,0,0],dtype=int),['negative_shift','id']],         
            [np.array([0,0,1],dtype=int),['inverse_fourier_transform','id','id']]
        ]
        shift_to_center_sketch = [
            [(0,0,1),['find_center','fourier_transform','id']],
            [(1,0,2,0,0),['negative_shift','negative_shift','id']],         
            [(0,1,2),['inverse_fourier_transform','id','id']]
        ]
        shift_to_center = self.process_factory.buildProcessFromSketch(shift_to_center_sketch)
        return shift_to_center
    def assemble_rotate_by(self):
        rotate_by_sketch = [
            ['complex_harmonic_transform','id'],
            'rotate',
            'complex_inverse_harmonic_transform',
            ['id','fourier_transform']
        ]
        rotate_by = self.process_factory.buildProcessFromSketch(rotate_by_sketch).run
        return rotate_by
    def assemble_shift_by(self):
        shift_by_sketch = [
            ['fourier_transform','id'],
            'shift',
            ['inverse_fourier_transform','id']
        ]
        shift_by = self.process_factory.buildProcessFromSketch(shift_by_sketch).run
        return shift_by
    
    def generate_alignment_loop(self):
        align = self.align
        max_iterations = self.opt['max_iterations']
        error_limit = self.opt['alignment_error_limit']
        integrate = mLib.SphericalIntegrator(self.ft_grids.realGrid[:]).integrate_normed
        shift = self.process_factory.get_operator('shift')
        ft = self.process_factory.get_operator('fourier_transform')
        ift = self.process_factory.get_operator('inverse_fourier_transform')
        self.results['shifts']=[np.array([0,0,0])]
            
        def alignment_loop(reference, signal):
            self.results['rotation_metrics']=[]
            self.results['rotation_angles']=[]
            ref = reference.copy()
            sig = signal[0]
            ft_sig = signal[1]
            ref_norm = integrate(ref.real**2)
            if ref_norm == 0:
                ref_norm = 1
            errors = []
            for i in range(max_iterations):
                sig,ft_sig = align(ref,sig,ft_sig)
                
                #calc error
                diff = ref.real - sig.real
                error = integrate(diff**2)/ref_norm
                errors.append(error)
                last_shift = self.results['shifts'][-1]
                #log.info('\n error = {} last_shift = {} \n'.format(error,last_shift))
                # brake the loop if shift is 0 or error limit is reached
                if (error<error_limit) or (last_shift[0]==0):
                    break

            if False: #last_shift[0] != 0:
                # calculate total shift of reference
                cart_shift = np.sum(np.array([mLib.spherical_to_cartesian(shift) for shift in self.results['shifts']]),axis = 0)
                # apply corresponding shift to signal            
                shift = mLib.cartesian_to_spherical(-cart_shift)
                sig = self.shift_by(sig,shift) #returns (density, ft(density))
            else:
                sig = [sig,ft_sig]
            out = {'densities':sig,'errors':errors,'shifts':self.results['shifts'],'rotation_metrics':self.results['rotation_metrics'],'rotation_angles':self.results['rotation_angles']}
            return out
        def alignment_routine(reference,signal):
            #log.info("type {}".format(type(signal)))
            #log.info("len {}".format(len(signal)))            
            #signal_inverted = (ift(signal[1].conj()),signal[1].conj())
            signal_inverted = (ift(ft(signal[0]).conj()),signal[1].conj())
            
            out = alignment_loop(reference,signal)
            out_inverted = alignment_loop(reference,signal_inverted)
            reference_norm = integrate(reference.real**2)
            diff = reference.real - out['densities'][0].real
            diff_inverted = reference.real - out_inverted['densities'][0].real
            error = integrate(diff**2)/reference_norm
            error_inverted = integrate(diff_inverted**2)/reference_norm
            #log.info('error = {} inverted error ={}'.format(error,error_inverted))
            if error < error_inverted:
                return_dict = out
            else:
                return_dict = out_inverted
            #log.info('normal error = {} inverted error = {},signal max = {} reference max = {}'.format(error,error_inverted,signal[0].real.max(),reference[0].real.max()))
            return_dict['diff_norm']=diff
            return_dict['diff_inverted_norm']=diff_inverted
            return return_dict
        return alignment_routine



import numpy as np
import os
from os import path as op
import time
import logging
import traceback
import re
import glob


from xframe.interfaces import DatabaseInterface
from xframe.library.mathLibrary import plane3D
from xframe.library.mathLibrary import spherical_to_cartesian
from xframe.library.gridLibrary import GridFactory
from xframe.library.gridLibrary import NestedArray
from xframe.library.gridLibrary import double_first_dimension
from xframe.library.pythonLibrary import convert_to_slice_if_possible
from xframe.library.pythonLibrary import split_into_simple_slices
from xframe.library.pythonLibrary import split_ids_by_unique_values
from xframe.library.pythonLibrary import getArrayOfArray
from xframe.library.pythonLibrary import xprint
from xframe.database.database import DefaultDB
from xframe import Multiprocessing

log = logging.getLogger("root")


class ExperimentDB(DefaultDB,DatabaseInterface):
    def __init__(self,**folders_files):
        super().__init__(**folders_files)
        self.data_path = self.get_path('vds_agipd_data',is_file=False)
        self.mask_path = self.get_path('vds_agipd_mask',is_file=False)
        self.gain_path = self.get_path('vds_agipd_gain',is_file=False)
        self.train_path = self.get_path('vds_agipd_trainId',is_file=False)
        self.pulse_path = self.get_path('vds_agipd_pulseId',is_file=False)
        self.cell_path = self.get_path('vds_agipd_cellId',is_file=False)
        self.baseline_shift_path = self.get_path('vds_agipd_baseline_shift',is_file=False)
        self.frame_mask_path = self.get_path('vds_agipd_frame_mask',is_file=False)
        self.pump_diod_path = self.get_path('h5_pump_diod',is_file=False)
        self.comm_module = Multiprocessing.comm_module

        self.h5_VirtualSource = self.get_db("file://_.h5").h5.VirtualSource
        self.h5_VirtualLayout = self.get_db("file://_.h5").h5.VirtualLayout


    ## load_default, save_default will be depricated use save_direct,load_direct inherited from default_DB
    def load_default(self,name,**kwargs):
        return super().load(name,skip_custom_methods=True,**kwargs)
    def save_default(self,name,data,**kwargs):
        super().save(name,data,skip_custom_methods=True,**kwargs)
        

    ## general methods
    def load_vds(self,name,**opt):
        '''Custom load routine for agipd vds files. If vds files exest load them otherwise if options allow vds creation create vds and load it.'''
        path_modifiers = opt['path_modifiers']
        file_path = self.get_path('vds',path_modifiers = path_modifiers)
        #log.info(file_path)
        if op.exists(file_path):
            data = self.load_direct('vds',**opt)
        else:
            log.info('vds_file: {} \n does not exist yet.'.format(file_path))
            if opt['allow_vds_creation']:
                log.info('creating vds file. data_mode = {}'.format(path_modifiers['data_mode']))
                run = path_modifiers['run']
                from_raw_data = (path_modifiers['data_mode'] == 'raw')                
                self.create_vds(run, from_raw_data,n_processes=opt.get('n_processes',False))
                log.info('finished vds creation.')
                data = self.load_direct('vds',**opt)
        #log.info('finished lading for {}.  exiting load vds'.format(path_modifiers['data_mode']))
        return data
    def _load_vds_old(self,name,**opt):
        path_modifiers = opt['path_modifiers']
        file_path = self.get_path('vds',path_modifiers = path_modifiers)
        if op.exists(file_path):
            data = self.load_default('vds',**opt)
        else:
            log.info('vds_file: {} \n does not exist.'.format(file_path))
            if opt['allow_vds_creation']:
                log.info('creating vds file. mode = {}'.format(path_modifiers['data_mode']))
                run = path_modifiers['run']
                from_raw_data = (path_modifiers['data_mode'] == 'raw')
                self.create_vds(run, from_raw_data,create_modulewise_vds = opt.get('create_modulewise_vds',False),create_complete_vds=True,n_processes=opt.get('n_processes',False))
                log.info('finished vds creation call load_default')
                data = self.load_default('vds',**opt)
        log.info('finished lading for {}.  exiting load vds'.format(path_modifiers['data_mode']))
        return data
    
    def load_info(self,name,**opt):
        runs = opt['runs'] #int or iterable
        data_mode = opt['data_mode']
        
        if isinstance(runs,int):
            runs = (runs,)
        infos = {}
        for run in runs:
            #log.info('type of run = {}'.format(run))
            with self.load('vds',path_modifiers={'run':run,'data_mode':data_mode,'module':0},as_h5_object = True) as h5_file:
                #log.info('opened {} file'.format(data_mode))
                ## loads general information about the dataset assumes modules,n_patterns,x_pixel_axis(long),y_pixel_axis(short) data shape
                #valid_frames_mask = self._get_valid_frames_mask(h5_file)
                data_set = h5_file[self.data_path]
                shape = data_set.shape
                infos[str(run)]={'n_frames': shape[1],
                                 'frame_size_Bytes': np.prod(data_set.shape[2:])*data_set.dtype.itemsize,
                                 'data_shape':shape,
                                 'cells': np.unique(h5_file[self.cell_path][:1000])
                }
            #log.info('finished generating info.')
        return infos

    def load_cell_ids(self,name, **opt):
        run = opt['path_modifiers']['run']
        data_mode = opt['path_modifiers']['data_mode']
        
        frames = opt.get('frames',slice(None))
        with self.load('vds',path_modifiers={'run':run,'data_mode':data_mode,'module':0},as_h5_object = True) as h5_file:
            data=h5_file[self.cell_path][frames]#frame,cell_id
        return data
    def load_pulse_ids(self,name, **opt):
        run = opt['path_modifiers']['run']
        data_mode = opt['path_modifiers']['data_mode']
                
        frames = opt.get('frames',slice(None))
        with self.load('vds',path_modifiers={'run':run,'data_mode':data_mode,'module':0},as_h5_object = True) as h5_file:
            data=h5_file[self.pulse_path][frames]#frame,pulse_id
        return data
    def load_train_ids(self,name, **opt):
        run = opt['path_modifiers']['run']
        data_mode = opt['path_modifiers']['data_mode']
                
        frames = opt.get('frames',slice(None))
        with self.load('vds',path_modifiers={'run':run,'data_mode':data_mode,'module':0},as_h5_object = True) as h5_file:
            data=h5_file[self.train_path][frames]#frame,train_id
        return data
    
    def load_baseline_shift(self,name, **opt):
        run = opt['path_modifiers']['run']
        data_mode = opt['path_modifiers']['data_mode']
                
        frames = opt.get('frames',slice(None))
        with self.load('vds',path_modifiers={'run':run,'data_mode':data_mode,'module':0},as_h5_object = True) as h5_file:
            data=h5_file[self.baseline_shift_path][frames]#frame,train_id
        return data
    def load_frame_mask(self,name, **opt):
        run = opt['path_modifiers']['run']
        data_mode = opt['path_modifiers']['data_mode']
                
        frames = opt.get('frames',slice(None))
        with self.load('vds',path_modifiers={'run':run,'data_mode':data_mode,'module':0},as_h5_object = True) as h5_file:
            data=h5_file[self.frame_mask_path][frames]#frame,train_id
        return data
    
    def load_custom_mask(self,name,**opt):
        mask_dict = super().load(name,skip_custom_methods=True,path_modifiers={'name':opt['file_name']})
        return mask_dict[opt['key']].astype(bool)
    
    def load_geometry(self,name,**kwargs):
        modules = range(16)
        geom_path = super().get_path('geometry')
        file_type = os.path.splitext(geom_path)[1] 
        #log.info('geom_cryst_path = {}'.format(file_type))
        if os.path.isfile(geom_path):
            if file_type == '.h5':
                plane_dict = super().load('geom')['modules']
                planes = tuple( plane3D.from_array(plane_dict[str(module)]) for module in modules)
            elif file_type == '.geom':
                #log.info('yay')
                search_phrases=tuple('p{}a0/fs = '.format(module) for module in range(16))
                with open(geom_path, "r") as geom:
                    lines = geom.readlines()
                r_float = '[\-]*[0-9]+\.[0-9]*'
                r_module = '[0-9]+[a-z]'

                data={}
                for id,line in enumerate(lines):
                    if any(tuple(phrase in line for phrase in search_phrases)):
                        module = re.search(r_module,line)[0][:-1]
                        y_direction = np.array(re.findall(r_float,line)+['0.0']).astype(float)
                        x_direction = np.array(re.findall(r_float,lines[id+1])+['0.0']).astype(float)
                        base = np.array(re.findall(r_float,lines[id+2])+re.findall(r_float,lines[id+3])+['0.0']).astype(float)*0.2 #0.2mm pixelwidth | geom file etries are measured in pixel width
                        data[module]=plane3D(base=base,x_direction =x_direction,y_direction = y_direction)

                planes= tuple(data[str(module)] for module in modules)
            else:
                raise IOError('Geometry file type {} not .h5 or .geom'.format(file_type))
        else:
            raise IOError('Geometry file not found at {}'.format(geom_path))
        return planes

    
    ## select specific frame range and split it into chunks  ##
    def _process_frame_range(self,frame_range):
        'if its a string threat it as datapath to some fole that contains a frame range'
        if isinstance(frame_range,str):
            frame_range = self.load(frame_range)
        return frame_range

    def get_frame_ids_from_selection(self,frame_range,selection,good_cells,h5_file):
        n_frames = len(h5_file[self.cell_path])
        frame_mask = np.zeros(n_frames,dtype = bool)
        frame_mask[frame_range] = True 
        
        dsets = {}
        u_dsets = {}
        
        dsets['cells'] = h5_file[self.cell_path][:]
        dsets['trains'] = h5_file[self.train_path][:]
        dsets['pulses'] = h5_file[self.pulse_path][:]


        bad_cell_mask = np.isin(dsets['cells'],good_cells)
        sel_mask = bad_cell_mask
        for key in dsets:
            sel = selection[key]
            sel_mask &= sel.mask(dsets[key])
            
        mask = sel_mask & frame_mask
        #mask[:frame_range.start] = False
        #mask[frame_range.stop:] = False
        frame_ids = np.nonzero(mask)[0]
        n_frames = len(frame_ids)
        return frame_ids,n_frames

    def load_chunks(self, name ,**opt):
        '''
        Gets the frame Ids corresponding to kwargs['selection'] and splits them in chunks according to the available memory.
        '''
        run = opt['run']
        data_mode = opt['data_mode']
        frame_range = self._process_frame_range(opt['frame_range'])
        #log.info('frame_range = {}'.format(frame_range))
        selection = opt['selection']
        modules = opt['modules']
        n_out_frames = opt['n_frames']
        free_mem = opt['free_mem']
        good_cells = opt['good_cells']
        in_multiples_of = opt['in_multiples_of']
        
        with self.load('vds',path_modifiers={'run':run,'data_mode':data_mode,'module':0},as_h5_object = True) as h5_file:            
            frame_ids,n_frames = self.get_frame_ids_from_selection(frame_range,selection,good_cells,h5_file)
            if not isinstance(n_out_frames,bool):                
                try:
                    assert (n_frames >= n_out_frames), 'Not enough frames beeing selected. Required {} frames got {} frames. Try to continue.'.format(n_out_frames,n_frames)
                except AssertionError as e:
                    log.error(e)
                    n_out_frames = n_frames
                    #log.info('n_frames {} nout frames {}'.format(n_frames,n_out_frames))
                    #raise e
                frame_ids = frame_ids[:n_out_frames]#np.random.choice(frame_ids,size = n_out_frames, replace=False)
                #log.info('n frames after = {}'.format(len(frame_ids)))
                n_frames= n_out_frames            
            #log.info('frame_ids shape ={}'.format(frame_ids.shape))
            dataset = h5_file[self.data_path]
            frame_shape = dataset.shape[1:]
            #log.info('frame shape = {}'.format(frame_shape))
            frame_size_B = np.prod(frame_shape)*len(modules)*np.dtype(float).itemsize
            #log.info('frame size = {} [MBytes]'.format(frame_size_B/1024**2))
        simultanous_frames_in_mem = 1 + free_mem//frame_size_B
        #log.info('simultaneouse frames in mem = {} [MBytes]'.format(simultanous_frames_in_mem))
        #make sure all but the last chunk part are in multiples of 'in_multiples_of'
        #log.info('in multiples of = {}'.format(in_multiples_of))
        if not isinstance(in_multiples_of,bool):
            simultanous_frames_in_mem = simultanous_frames_in_mem//in_multiples_of *in_multiples_of
            #log.info('simultanous_frames_in_mem = {}'.format(simultanous_frames_in_mem))
        #log.info('n_frames = {}'.format(n_frames))
        split_ids = np.concatenate((np.arange(0,n_frames,simultanous_frames_in_mem),[n_frames-1])).astype(int)
        #log.info('split_ids={}'.format(split_ids))
        #if last chunk part is smaller than 'in_multiples_of' combine the last two chunk parts.
        #i.e make sure that each chunk is bigger than in_multiples_of
        if not isinstance(in_multiples_of,bool):
            last_chunk_size = n_frames%simultanous_frames_in_mem
            if last_chunk_size < in_multiples_of:
                split_ids=split_ids[:-1]
                
        log.info("Maximal chunk size is {} GB or {} frames".format(simultanous_frames_in_mem*frame_size_B/1024**3,simultanous_frames_in_mem))

        if len(frame_ids) !=0:
            chunks=np.split(frame_ids,split_ids[1:-1])
        else:
            chunks = []
        #log.info('n frames = {}'.format(len(frame_ids)))
        #log.info('frames in chunks = {}'.format(np.sum([len(chunk) for chunk in chunks])))
        return chunks
            

    ## load data from VDS files 
    def _load_data_chunk_worker(self,modules,slices,module_id_lookup,run,data_slices,output_slices,data_mode,**opt):
        for m,s_id in zip(modules,slices):
            #log.info('mdule {}'.format(m))
            m_id = module_id_lookup[m]
            data,mask,gain = opt['outputs'][:3]
            d_slice = data_slices[s_id]
            o_slice = output_slices[s_id]
            with super().load('vds',path_modifiers={'run':run,'data_mode':data_mode,'module':m},as_h5_object = True) as h5_file:                
                vds = h5_file[self.data_path]
                vds_mask = h5_file[self.mask_path]
                vds_gain = h5_file[self.gain_path]                
                #for d_slice,o_slice in zip(data_slices,output_slices):
                #log.info('out_slice shape = {}'.format(o_slice))
                #log.info('data_slice shape = {}'.format(vds[d_slice].shape))
                vds.read_direct(data,d_slice,(m_id,o_slice))
                nan_mask = np.isnan(vds[d_slice])
                if nan_mask.any():
                    log.info('module {}, slice {}, out_slice {}'.format(m,d_slice,o_slice))
                    log.info('{}% nans found '.format(np.sum(nan_mask)/np.prod(nan_mask.shape)*100))
                if data_mode =='proc':
                    vds_mask.read_direct(mask,d_slice,(m_id,o_slice))
                    vds_gain.read_direct(gain,d_slice,(m_id,o_slice))
                # inverting the mask so that unmasked values are 1/True and masked values are 0/False
                mask[m_id,o_slice] = ~mask[m_id,o_slice]
    def _load_data_chunk_worker2(self,module,module_id_lookup,run,data_slices,output_slices,data_mode,**opt):
        m = module
        for d_slice,o_slice in zip(data_slices,output_slices):
            #log.info('mdule {}'.format(m))
            m_id = module_id_lookup[m]
            data,mask,gain = opt['outputs'][:3]
            #d_slice = data_slices[s_id]
            #o_slice = output_slices[s_id]
            with super().load('vds',path_modifiers={'run':run,'data_mode':data_mode,'module':m},as_h5_object = True) as h5_file:
                vds = h5_file[self.data_path]
                vds_mask = h5_file[self.mask_path]
                vds_gain = h5_file[self.gain_path]                
                #for d_slice,o_slice in zip(data_slices,output_slices):
                #log.info('out_slice shape = {}'.format(o_slice))
                #log.info('data_slice shape = {}'.format(vds[d_slice].shape))
                vds.read_direct(data,d_slice,(m_id,o_slice))
                nan_mask = np.isnan(vds[d_slice])
                if nan_mask.any():
                    log.info('module {}, slice {}, out_slice {}'.format(m,d_slice,o_slice))
                    log.info('{}% nans found '.format(np.sum(nan_mask)/np.prod(nan_mask.shape)*100))
                if data_mode =='proc':
                    vds_mask.read_direct(mask,d_slice,(m_id,o_slice))
                    vds_gain.read_direct(gain,d_slice,(m_id,o_slice))
                # inverting the mask so that unmasked values are 1/True and masked values are 0/False
                mask[m_id,o_slice] = ~mask[m_id,o_slice]                
    def load_data_chunk(self,name, **opt):
        run = opt['run']
        modules = opt['modules']
        frame_ids = opt['frame_ids']
        data_mode = opt['data_mode']
        n_processes = opt.get('n_processes',False)
        #outputs = opt['outputs']        
        
        n_frames = len(frame_ids)
        with self.load('vds',path_modifiers={'run':run,'data_mode':data_mode,'module':0},as_h5_object = True) as h5_file:
            chunk_shape = (len(modules),n_frames) + h5_file[self.data_path].shape[1:]
        
        frame_slices,out_slices = split_into_simple_slices(frame_ids,return_sliced_args=True)
        data,mask,gain = self.comm_module.request_mp_evaluation(self._load_data_chunk_worker , mp_type='shared_array_multi_new' , out_shapes = [chunk_shape,chunk_shape,chunk_shape], out_dtypes = [np.dtype('float32'),np.dtype('bool'),np.dtype('uint8')],argArrays = [modules,np.arange(len(frame_slices))], const_args = [run,frame_slices,out_slices,data_mode], callWithMultipleArguments = True,splitMode='modulus',n_processes = n_processes)
        data = np.swapaxes(data,0,1)
        if data_mode == 'proc':
            mask = np.swapaxes(mask,0,1)
            gain = np.swapaxes(gain,0,1)
            return data,mask,gain
        else:
            return data


    ## HDF5 VDS Generation ##
    def create_vds(self, run, from_raw_data,modules = np.arange(16), n_processes=False):
        ## There is a multiprocessing bug in the following code right now only works of one module is given as input
        '''
        This function creates assembled vds files for data from the AGIPD detector at the European XFEL.
        It supports to generate sepparate vds per detector module or a combined vds for all modules (this is however slower).
        It automatically filters out trainId==0, doubled train_ids and pulseId or cellId == 65535
        input:
            run int : The Experimental run to generatevds files for
            from_raw_data bool: Whether to generate vds_data for raw or proc files
            create_modulewise_vds bool: Whether or not to create module wise vds files
            create_complete_vds bool: Whether or not to create fully combined vds_files
            n_processes bool or int: If bool it uses the non-hyperthreading threads on your machine (i.e. number of cpu cores). 
        '''
        comm_module = Multiprocessing.comm_module
        if from_raw_data:
            data_mode = 'raw'
        else:
            data_mode = 'proc'
            
        ## get file path
        #paths = tuple( sorted(glob.glob(folder+'/*AGIPD%.2d*.h5'%module))for module in range(16) )
        xprint(f"Combining AGIPD data for run {run} into virtual h5 datasets.")
        paths = [sorted(glob.glob(self.get_path('vds_regexpr',path_modifiers = {'data_mode':data_mode,'run':run,'module':module}))) for module in modules]
        m_ids = np.concatenate([np.full(len(module_paths),m) for m,module_paths in zip(modules,paths)])
        #m_ids = np.concatenate(tuple(np.full(len(file_list),module,dtype = np.int) for module,file_list in enumerate(paths)))        
        #flist = np.array( tuple( path for paths_per_module in paths for path in paths_per_module))
        flist = np.array(tuple(path for module_paths in paths for path in module_paths))
        assert len(flist)>0,'No input files found at {} check file "vds_regexpr" in experimental settings.'.format(self.get_path('run_data',is_file=False,path_modifiers = {'data_mode':data_mode,'run':run}))
        #log.info(tuple(zip(m_ids,flist)))
        ## setup general information about datasets
        log.info('Loading general dataset info.')
        general_data = self.get_general_vds_data(flist,m_ids,from_raw_data,n_processes = n_processes)        
            
        ## Count number of frames and create a way to find the part of frames corresponding to a given file.     
        lookup_array = general_data['train_lookup_array']
        start_frames = [np.sum(lookup_array[:i,0]) for i in range(len(lookup_array)+1)]
        frame_slices = [slice(start_frames[i],start_frames[i+1]) for i in range(len(lookup_array))]        
        n_frames = general_data['n_frames']
        frame_ids = np.arange(n_frames)        
        log.info('Generating VDS layouts.')
        mp_mode = Multiprocessing.MPMode_Queue(assemble_outputs = False)
        results = comm_module.request_mp_evaluation(self._vds_worker,mp_mode,input_arrays=[flist,m_ids],const_inputs=[frame_slices,frame_ids,lookup_array,general_data,data_mode],split_together = True,n_processes = n_processes)
        #log.info('len(m_ids) = {}'.format(len(m_ids)))
        results = tuple( elem for result in results.values() for elem in result)
        
        result_module_ids = np.array([r['module'] for r in results])
        log.info('Populating VDS layouts and saving VDS files.')
        comm_module.request_mp_evaluation(self._generate_module_wise_vds_files,mp_mode,input_arrays=[modules],const_inputs = [run,results,result_module_ids,frame_slices,general_data,from_raw_data],n_processes = n_processes)
        xprint("done.")
        
    def get_general_vds_data(self,flist,m_ids,module,from_raw_data= False,n_processes =False):
        if from_raw_data:
            data_mode = 'raw'
        else:
            data_mode = 'proc'
        flist = np.asarray(flist)
        def _extract_general_data_worker(file_names,module_ids,**kwargs):
            n_frames=[]
            n_trains = []
            last_train_ids = []
            max_train_ids = []
            first_train_ids = []
            for fname,m_id in zip(file_names,module_ids):
                train_path = self.get_path('h5_agipd_trainId',is_file = False,path_modifiers = {'module':m_id})
                pulse_path = self.get_path('h5_agipd_pulseId',is_file = False,path_modifiers = {'module':m_id})
                data_path = self.get_path('h5_agipd_data',is_file = False,path_modifiers = {'module':m_id})
                with self.load_direct(fname, as_h5_object=True) as f:
                    n_frames.append(f[data_path].shape[0])
                    train_ids = f[train_path][:]
                    min_train_id = int(np.min(train_ids[train_ids>0]))
                    unique_pulses,counts = np.unique(f[pulse_path][:],return_counts = True)
                    train_count = int(np.median(counts))
                    n_trains.append(train_count)
                    first_train_ids.append(min_train_id)
                    last_train_ids.append(min_train_id+train_count-1)
                    max_train_ids.append(np.max(train_ids))
                    #log.info(f'module {m_id} n_frames: {n_frames[-1]} file name = {fname}')
            #log.info(f'outputs = { np.array([n_frames,first_train_ids,last_train_ids,max_train_ids,n_trains,module_ids])}')
            return [n_frames,first_train_ids,last_train_ids,max_train_ids,n_trains,module_ids]
        mp_mode = Multiprocessing.MPMode_Queue(assemble_outputs = False)
        results2 = Multiprocessing.comm_module.request_mp_evaluation(_extract_general_data_worker,mp_mode,input_arrays = [flist,m_ids],split_together=True, call_with_multiple_arguments = True,n_processes = n_processes)
        #log.info('{}'.format([np.asarray(a).T.shape for a in results.values()]))
        #log.info('results shape = {}'.format(results.shape))
        results2 = np.concatenate([np.asarray(a).T for a in results2.values()],axis = 0)

        results = Multiprocessing.comm_module.request_mp_evaluation(_extract_general_data_worker,input_arrays = [flist,m_ids],split_together=True, call_with_multiple_arguments = True,n_processes = n_processes)
        results = np.concatenate([np.asarray(a).T for a in results],axis = 0)
        #log.info(f'{[len(a) for a in results2]}, len flist = {len(flist)} len results = {len(results2)} len mids = {len(m_ids)}')

        #Split results by module
        #args_sorted_by_modules = np.argsort(results[:,-1])
        #results = results[args_sorted_by_modules]
        results_per_module=np.array(np.split(results,split_ids_by_unique_values(results[:,-1])))
        
        log.info(f'{np.array(results_per_module)[:,:,0]}')
        #Check whether all modules have same number of frames and trains
        n_frames_per_module = results_per_module[:,:,0].sum(axis = 1)
        n_trains_per_module = results_per_module[:,:,4].sum(axis = 1)
        #nnumber_of_frames_equal = [np.sum(r[:,4]) for r in results_per_module]
        all_modules_have_same_number_of_frames = not ((n_frames_per_module-n_frames_per_module[0]).any())
        all_modules_have_same_number_of_trains = not ((n_trains_per_module-n_trains_per_module[0]).any())                
        #assert all_modules_have_same_number_of_trains,'Not all modules have same number of trains. \n{}'.format(n_trains_per_module)
        try:
            assert all_modules_have_same_number_of_frames,'Not all modules have same number of frames. \n{}'.format(n_frames_per_module)
        except AssertionError as e:
            log.error(e)
            meta_data_equal = (results_per_module[:,:,0] != results_per_module[0,None,:,0])
            differences_in = np.stack(np.nonzero(meta_data_equal),axis=-1) 
            log.info(f'Differences in modules/files {differences_in}')
            raise e
        

        #From now on consider only results of 0 module and remove the module id data 
        results = results_per_module[0][:,:-1]
        
        #log.info('results shape = {}'.format(results.shape))
        train_lookup_array = results[np.argsort(results[:,2])][:,:3].astype(int)
        #log.info('n_frames = {}'.format(train_lookup_array[:,0]))
        #log.info('lookup_array = {}'.format(train_lookup_array))
        n_frames = int(np.sum(results[:,0]))
        n_trains = int(np.sum(results[:,-1]))
        first_train = int(np.min(results[:,1]))
        last_train = int(first_train + n_trains -1)
        max_train_id = int(np.max(results[:,3]))
        
        out_dict = {'n_frames': n_frames,
                    'n_trains': n_trains,
                    'first_train_id':first_train,
                    'last_train_id':last_train,
                    'max_train_id':max_train_id,
                    'train_lookup_array':train_lookup_array
                    }
        #log.info('first train = {}, last train = {} ,n_trains = {}, max_train = {}'.format(first_train,last_train,n_trains,max_train_id))
        data_path = self.get_path('h5_agipd_data',is_file = False,path_modifiers = {'module':m_ids[0]})
        cell_path = self.get_path('h5_agipd_cellId',is_file = False,path_modifiers = {'module':m_ids[0]})
        train_path = self.get_path('h5_agipd_trainId',is_file = False,path_modifiers = {'module':m_ids[0]})
        pulse_path = self.get_path('h5_agipd_pulseId',is_file = False,path_modifiers = {'module':m_ids[0]})
        with self.load_direct(flist[0],as_h5_object=True) as f: #h5.File(flist[0], 'r') as f:
            #log.info(f['INSTRUMENT/'+det_name +'/DET/'].keys())
            data = f[data_path]                
            data_shape = data.shape
            pulses = f[pulse_path][:]
            cells = f[cell_path]
            trains = f[train_path]
            uniq_pulses = np.unique(pulses)
            out_dict['data_shape'] = (n_frames,)+data_shape[1:]
            out_dict['uniq_pulses'] = uniq_pulses
            out_dict['n_pulses']=len(uniq_pulses)
            out_dict['data_dtype']=data.dtype
            out_dict['pulse_dtype']=pulses.dtype
            out_dict['cell_dtype']=cells.dtype
            out_dict['train_dtype']=trains.dtype
            if not from_raw_data:
                mask_path = self.get_path('h5_agipd_mask',is_file = False,path_modifiers = {'module':m_ids[0]})
                gain_path = self.get_path('h5_agipd_gain',is_file = False,path_modifiers = {'module':m_ids[0]})
                blShift_path = self.get_path('h5_agipd_baseline_shift',is_file = False,path_modifiers = {'module':m_ids[0]})
                #log.info(f['INSTRUMENT/'+det_name +'/DET/{}CH0:xtdf/image/'.format(module)].keys())
                gain = f[gain_path]
                mask = f[mask_path]
                baseline_shift = f[blShift_path]
                out_dict['mask_dtype'] = mask.dtype
                out_dict['gain_dtype'] = gain.dtype
                out_dict['basline_shift_dtype'] = baseline_shift.dtype
        return out_dict 
        
    def _get_global_slice_id(self,train_id,lookup_array):
        slice_id = np.argmin(lookup_array[:,1]<=train_id)-1
        #log.info(slice_id)
        #log.info(frame_slices[slice_id])
        if (train_id>lookup_array[-1,2]) or (train_id<lookup_array[0,1]):
            raise AssertionError('train_id {} outside of known train id values [{},{}]'.format(train_id,lookup_array[0,1],lookup_array[-1,2]))
        else:
            return slice_id
    def _vds_worker(self,file_name,module,frame_slices,frame_ids,slice_lookup_array,general_data,data_mode,**kwargs):
        data_path = self.get_path('h5_agipd_data',is_file = False,path_modifiers = {'module':module})
        train_path = self.get_path('h5_agipd_trainId',is_file = False,path_modifiers = {'module':module})
        pulse_path = self.get_path('h5_agipd_pulseId',is_file = False,path_modifiers = {'module':module})
        cell_path = self.get_path('h5_agipd_cellId',is_file = False,path_modifiers = {'module':module})
        with self.load_direct(file_name, as_h5_object=True) as f: #h5.File(file_name, 'r') as f:
            #log.info(f[train_path][:])
            # Load 1D datasets of traint/pulse/cell ids which will be combined and stored directly without virtualization
            # Annoyingly, raw data has an extra dimension for the IDs
            # (which is why we need the ravel)
            train_ids = f[train_path][:].ravel() # equivalent to  f[train_path][:].reshape(-1)
            cell_ids = f[cell_path][:].ravel() # equivalent to  f[cell_path][:].reshape(-1)
            pulse_ids = f[pulse_path][:].ravel() # equivalent to  f[pulse_path][:].reshape(-1)
            part_id = self._get_global_slice_id(train_ids[0],slice_lookup_array)
            #log.info('module = {} train_id = {} part_id ={}'.format(module,train_ids[0],part_id))
            part_slice = frame_slices[part_id]
            frame_ids_part = frame_ids[part_slice]
                
            # Generate frame mask by excluding bad trains cells and pulses:                
            # Remove the following bad data:
            #   Train ID = 0, suggesting no input from AGIPD
            #   Train ID out of range, for bit flips from the trainID server
            #   Repeated train IDs: Keep only first train with that ID
            first_global_train_id = general_data['first_train_id']
            last_global_train_id = general_data['last_train_id']
            n_pulses =  general_data['n_pulses']
            train_mask = (train_ids>0) & (train_ids>=first_global_train_id) & (train_ids<=last_global_train_id)
            uniq_trains, uniq_trains_counts = np.unique(train_ids, return_counts=True, return_index=True)[1:]
            for i in uniq_trains[uniq_trains_counts > n_pulses]:
                log.warning('WARNING: Repeated train IDs in {} from ind {}'.format((op.basename(file_name), i)))
                train_mask[np.nonzero(train_ids==train_ids[i])[0][n_uniq_pulses:]] = False                 
            local_frame_mask = train_mask

            if (~local_frame_mask).any():
                log.info('{} of {} trains are masked.'.format(np.sum(~local_frame_mask),uniq_trains.shape))
                log.info("first train = {} last train = {} train_ids:\n {} ".format(first_global_train_id,last_global_train_id,train_ids))
            
            # Setting masked Ids to default values
            # pulse and cellIds are stored as uint64 dtype while the cell ids are stored as uint16
            # We set the default value to be the maximal possible integer for the given datatypes which are:
            # 2**64-1 for uint64 and 2**16-1 for uint16
            bad_data_mask = ~local_frame_mask
            max_uint64 = int(2**64-1)
            max_uint16 = int(2**16-1)
            train_ids[bad_data_mask]=max_uint64
            pulse_ids[bad_data_mask]=max_uint64
            cell_ids[bad_data_mask]=max_uint16
            #good_local_frame_ids = np.arange(len(train_ids))[local_frame_mask]
            #good_global_frame_ids = frame_ids_part[local_frame_mask]
            #n_good_frames = len(good_global_frame_ids)
            
            # Initialize results
            result = {
                'module':module,
                'cells':cell_ids,
                'pulses':pulse_ids,
                'trains':train_ids,
                'frame_ids':frame_ids_part,
                'local_frame_mask':local_frame_mask,
                'part_id': part_id,
            #    'n_good_frames':n_good_frames
            }
            
            # Load further 1D datasets that will be stored directly
            # baseline shift: Set masked valus to np.nan snce the data is of type float32
            if data_mode == 'proc':
                blShift_path = self.get_path('h5_agipd_baseline_shift',is_file = False,path_modifiers = {'module':module})
                baseline_shift = f[blShift_path][:]
                baseline_shift[bad_data_mask]=np.nan
                result['baseline_shift']=baseline_shift
                
            # Create VDS virtualsource data for 2d virtual datasets 
            data = f[data_path]
            vsource_data = self.h5_VirtualSource(data)
            result['v_source'] = vsource_data
            if data_mode == 'proc':
                mask_path = self.get_path('h5_agipd_mask',is_file = False,path_modifiers = {'module':module})
                gain_path = self.get_path('h5_agipd_gain',is_file = False,path_modifiers = {'module':module})             
                mask = f[mask_path]
                vsource_mask = self.h5_VirtualSource(mask)                    
                gain = f[gain_path]
                vsource_gain = self.h5_VirtualSource(gain)                
                result['v_source_mask'] = vsource_mask
                result['v_source_gain'] = vsource_gain
            return result
    def _generate_module_wise_vds_files(self,module_id,run,results,module_ids,frame_slices,general_data,from_raw_data,**args):
        if from_raw_data:
            data_mode = 'raw'
        else:
            data_mode = 'proc'
        result_ids = np.nonzero(module_ids==module_id)[0]
        #log.info('module_id = {}'.format(module_id))
        data_shape = general_data['data_shape']
        data_dtype = general_data['data_dtype']
        n_frames = data_shape[0]
        shape_1d = (n_frames,)
        layouts = {
            'data':self.h5_VirtualLayout(shape=data_shape, dtype = general_data['data_dtype']),
            # the following virtualLayouts are only used if data_mode = 'proc' in which case 'mask_dtype' and 'gain_dtype' are contained in general_data
            # I'm generating the following vds layouts also in case of data_mode = 'raw' with mock datatypes to eliminate if statements in this routine
            'mask':self.h5_VirtualLayout(shape=data_shape, dtype = general_data.get('mask_dtype',data_dtype)),
            'gain':self.h5_VirtualLayout(shape=data_shape, dtype = general_data.get('gain_dtype',data_dtype))
        }
        #log.info('data layout shape = {}'.format(layouts['data'].shape))
        cell_ids = np.zeros(shape_1d,dtype = general_data['cell_dtype'])
        pulse_ids = np.zeros(shape_1d,dtype= general_data['pulse_dtype'])
        train_ids = np.zeros(shape_1d,dtype= general_data['train_dtype'])
        frame_mask = np.zeros(shape_1d,dtype= bool)
            
        # Same comment as for the above as for gain and mask virtual layouts
        baseline_shifts = np.zeros(shape_1d,dtype = general_data.get('baseline_shift_dtype',data_dtype))
                             
        frame_mask = np.zeros(shape_1d,dtype= bool)
        for r_id in result_ids:
            result = results[r_id]
            part_id = result['part_id']
            frame_slice = frame_slices[part_id]
            #log.info('module {} partid = {}'.format(module_id,part_id))
            self._populate_vds(result,layouts,frame_slice,from_raw_data,index_module = False)
            cell_ids[frame_slice] = result['cells']
            pulse_ids[frame_slice] = result['pulses']
            train_ids[frame_slice] = result['trains']
            frame_mask[frame_slice] = result['local_frame_mask']
            baseline_shifts[frame_slice] = result.get('baseline_shift',0)
        
        vds_file_name = self.get_path('vds',path_modifiers={'run':run,'data_mode':data_mode,'module':module_id})
        #log.info('vds_file name = {}'.format(vds_file_name))
        self._save_vds_file(layouts,cell_ids,pulse_ids,train_ids,frame_mask,baseline_shifts,vds_file_name,from_raw_data)
        #log.info('done')            
        return None
    def _populate_vds(self,result,layouts,frame_slice,from_raw_data,index_module=True):
        module = result['module']
        vsource_data = result['v_source']
        layout_data = layouts['data']
        #log.info('layout_shape = {} vsource shape ={} slice = {} len slice = {}'.format(layout_data.shape,vsource_data.shape,frame_slice,frame_slice.stop-frame_slice.start))
        #log.info('slice in populate_vds ={}'.format(frame_slice))
        #log.info('max_local frame_id ={}'.format(local_frame_ids.max()))
        #log.info('layout  shape {}'.format(layout_data.shape))
        #log.info('vsource shape {}'.format(vsource_data.shape))             
        if index_module:
            layout_data[module,frame_slice] = vsource_data
        else:
            #log.info('vsource_shape = {}'.format(vsource_data.shape))
            #log.info('vsource_shape = {}'.format(vsource_data.shape))
            #log.info('layout_shape = {}'.format(layout_data[frame_slice].shape))
            #layout_data[frame_slice] = vsource_data[combined_good_frames_part]
            layout_data[frame_slice] = vsource_data
        if not from_raw_data:
            layout_mask = layouts['mask']
            vsource_mask = result['v_source_mask']                            
            layout_gain = layouts['gain']
            vsource_gain = result['v_source_gain']
            #log.info('layout gain shape {}'.format(layout_gain.shape))
            #log.info('vsource gain shape {}'.format(vsource_gain.shape))
            if index_module:
                layout_mask[module,frame_slice] = vsource_mask
                layout_gain[module,frame_slice] = vsource_gain
            else:
                #log.info('len frame_ids={}'.format(len(frame_ids)))
                layout_mask[frame_slice] = vsource_mask
                layout_gain[frame_slice] = vsource_gain
                
    def _save_vds_file(self,layouts,cell_ids,pulse_ids,train_ids,frame_mask,baseline_shifts,file_name,from_raw_data):
        data_path = self.get_path('vds_agipd_data',is_file = False)
        train_path = self.get_path('vds_agipd_trainId',is_file = False)
        pulse_path = self.get_path('vds_agipd_pulseId',is_file = False)
        cell_path = self.get_path('vds_agipd_cellId',is_file = False)
        
        frame_mask_path = self.get_path('vds_agipd_frame_mask',is_file = False)
        self.create_path_if_nonexistent(file_name)
        with self.save_direct(file_name,as_h5_object = True) as outf:
            outf.create_dataset(train_path,data = train_ids) 
            outf.create_dataset(cell_path,data = cell_ids)
            outf.create_dataset(pulse_path,data = pulse_ids)
            outf.create_dataset(frame_mask_path,data = frame_mask)
            outf.create_virtual_dataset(data_path, layouts['data'], fillvalue=np.nan)
            if not from_raw_data:
                mask_path = self.get_path('vds_agipd_mask',is_file = False)
                gain_path = self.get_path('vds_agipd_gain',is_file = False)
                baseline_shift_path = self.get_path('vds_agipd_baseline_shift',is_file = False)
                outf.create_virtual_dataset(mask_path, layouts['mask'], fillvalue=np.nan)
                outf.create_virtual_dataset(gain_path, layouts['gain'], fillvalue=np.nan)
                outf.create_dataset(baseline_shift_path,data = baseline_shifts)
            outf.close()
        

    def create_vds_module_old(self, run, from_raw_data,modules = np.arange(16), create_modulewise_vds = True, create_complete_vds = False, n_processes=False):
        ## There is a multiprocessing bug in the following code right now only works of one module is given as input
        '''
        This function creates assembled vds files for data from the AGIPD detector at the European XFEL.
        It supports to generate sepparate vds per detector module or a combined vds for all modules (this is however slower).
        It automatically filters out trainId==0, doubled train_ids and pulseId or cellId == 65535
        input:
            run int : The Experimental run to generatevds files for
            from_raw_data bool: Whether to generate vds_data for raw or proc files
            create_modulewise_vds bool: Whether or not to create module wise vds files
            create_complete_vds bool: Whether or not to create fully combined vds_files
            n_processes bool or int: If bool it uses the non-hyperthreading threads on your machine (i.e. number of cpu cores). 
        '''
        comm_module = Multiprocessing.comm_module
        if from_raw_data:
            data_mode = 'raw'
            folder = self.get_path('raw_data_base',path_modifiers={'run':run},is_file = False)
        else:
            data_mode = 'proc'
            folder = self.get_path('proc_data_base',path_modifiers={'run':run},is_file = False)
            
        ## get file path
        #paths = tuple( sorted(glob.glob(folder+'/*AGIPD%.2d*.h5'%module))for module in range(16) )
        paths = [sorted(glob.glob(folder+'/*AGIPD%.2d*.h5'%module)) for module in modules]
        m_ids = np.concatenate([np.full(len(module_paths),m) for m,module_paths in zip(modules,paths)])
        #m_ids = np.concatenate(tuple(np.full(len(file_list),module,dtype = np.int) for module,file_list in enumerate(paths)))        
        #flist = np.array( tuple( path for paths_per_module in paths for path in paths_per_module))
        flist = np.array(tuple(path for module_paths in paths for path in module_paths))
        #log.info(tuple(zip(m_ids,flist)))
        ## setup general information about datasets
        general_data = self.get_general_vds_data(flist[0],modules[0],from_raw_data)
        detector_name = general_data['detector_name']
        data_shape = general_data['data_shape']
        uniq_pulses = general_data['uniq_pulses']
        n_uniq_pulses = len(uniq_pulses)
        data_dtype = general_data['data_dtype']
        if not from_raw_data:
            mask_dtype = general_data['mask_dtype']
            gain_dtype = general_data['gain_dtype']
        cell_dtype = general_data['cell_dtype']
        pulse_dtype = general_data['pulse_dtype']
        train_dtype = general_data['train_dtype']
            
        ## Count number of frames and create a way to find the part of frames corresponding to a given file.     
        lookup_array = self.generate_frames_lookup_array(folder,n_processes = n_processes)
        last_train = lookup_array[-1,2]
        start_frames = [np.sum(lookup_array[:i,0]) for i in range(len(lookup_array)+1)]
        frame_slices = [slice(start_frames[i],start_frames[i+1]) for i in range(len(lookup_array))]
        def get_slice_id(train_id):
            slice_id = np.argmin(lookup_array[:,1]<=train_id)-1
            #log.info(slice_id)
            #log.info(frame_slices[slice_id])
            if (train_id>lookup_array[-1,2]) or (train_id<lookup_array[0,1]):
                raise AssertionError('train_id {} outside of known train id values [{},{}]'.format(train_id,lookup_array[0,1],lookup_array[-1,2]))
            return slice_id
        
        n_frames = np.sum(lookup_array[:,0])
        frame_ids = np.arange(n_frames)
        
        def vds_worker(file_name,module,**kwargs):
            dset_prefix = 'INSTRUMENT/'+detector_name+'/DET/%dCH0:xtdf/image/'%module
            with self.load_direct(file_name, as_h5_object=True) as f: #h5.File(file_name, 'r') as f:
                # Annoyingly, raw data has an extra dimension for the IDs
                # (which is why we need the ravel)
                train_ids = f[dset_prefix+'trainId'][:].ravel() # equivalent to  f[dset_prefix+'trainId'][:].reshape(-1)
                cell_ids = f[dset_prefix+'cellId'][:].ravel() # equivalent to  f[dset_prefix+'trainId'][:].reshape(-1)
                pulse_ids = f[dset_prefix+'pulseId'][:].ravel() # equivalent to  f[dset_prefix+'trainId'][:].reshape(-1)
                part_id = get_slice_id(train_ids[0])
                part_slice = frame_slices[part_id]
                frame_ids_part = frame_ids[part_slice]
                
                # Generate frame mask by excluding bad trains cells and pulses:                
                # Remove the following bad data:
                #   Train ID = 0, suggesting no input from AGIPD
                #   Train ID out of range, for bit flips from the trainID server
                #   Repeated train IDs: Keep only first train with that ID
                train_mask = (train_ids>0) & (train_ids<last_train) # gives true only for trIDs in the given range
                uniq_trains, uniq_trains_counts = np.unique(train_ids, return_counts=True, return_index=True)[1:]
                for i in uniq_trains[uniq_trains_counts > n_uniq_pulses]:
                    log.warning('WARNING: Repeated train IDs in {} from ind {}'.format((op.basename(file_name), i)))
                    train_mask[np.nonzero(train_ids==train_ids[i])[0][n_uniq_pulses:]] = False                 
                local_frame_mask = train_mask 
                good_local_frame_ids = np.arange(len(train_ids))[local_frame_mask]
                good_global_frame_ids = frame_ids_part[local_frame_mask]
                n_good_frames = len(good_global_frame_ids)
                
                # Initialize results
                result = {
                    'module':module,
                    'cells':cell_ids[local_frame_mask],
                    'pulses':pulse_ids[local_frame_mask],
                    'trains':train_ids[local_frame_mask],
                    'frame_ids':good_global_frame_ids,
                    'local_frame_mask':local_frame_mask,
                    'part_id': part_id,
                    'n_good_frames':n_good_frames
                }
                
                # Create VDS virtualsource data 
                data = f[dset_prefix+'data']
                vsource_data = self.h5_VirtualSource(data)
                result['v_source'] = vsource_data
                if not from_raw_data:
                    mask = f[dset_prefix+'mask']
                    vsource_mask = self.h5_VirtualSource(mask)                    
                    gain = f[dset_prefix+'gain']
                    vsource_gain = self.h5_VirtualSource(gain)                
                    result['v_source_mask'] = vsource_mask
                    result['v_source_gain'] = vsource_gain
                #if (data[0,0,0]<16.6) and (data[0,0,0]>16.5):
                #    log.info('found!!!!! data = {}'.format(data[0,0,0]))
                #    log.info('part_slice = {}'.format(part_slice))
                #    log.info('good_local_frame_ids = {}'.format(good_local_frame_ids[:10]))
                #    log.info('module = {}'.format(module))
                #    log.info('vsource_data ={}'.format(vsource_data[0,0,0]))
            return result
        def save_vds_file(layouts,cell_ids,pulse_ids,train_ids,n_good_frames,file_name):
            self.create_path_if_nonexistent(file_name)
            with self.save_direct(file_name, as_h5_object=True) as outf:
                #outf = h5.File(file_name, 'w', libver='latest')
                outf['INSTRUMENT/'+detector_name+'/DET/image/trainId'] = train_ids # create a dataset with all train numbers and write it to the h5file (considering pulse structure of the trains)
                # create datasets for cellid and pulseid and fill them with a number 65535
                default_value=65535
                outdset_cid = outf.create_dataset('INSTRUMENT/'+detector_name+'/DET/image/cellId',
                                                  shape=(n_good_frames,), dtype='u2',
                                                  data=np.full(n_good_frames,default_value, dtype=cell_dtype))
                outdset_cid[:]=cell_ids
                outdset_pid = outf.create_dataset('INSTRUMENT/'+detector_name+'/DET/image/pulseId',
                                                  shape=(n_good_frames,), dtype='u8',
                                                  data=np.full(n_good_frames,default_value, dtype=pulse_dtype))
                outdset_pid[:]=pulse_ids
                #assert (cell_ids != default_value).any(), 'There seem to be unset cell ids. Check vds_creation algorithm.'
                #assert (pulse_ids != default_value).any(), 'There seem to be unset pulse ids. Check vds_creation algorithm'
                outf.create_virtual_dataset('INSTRUMENT/'+detector_name+'/DET/image/data', layouts['data'], fillvalue=np.nan)
                if not from_raw_data:
                    outf.create_virtual_dataset('INSTRUMENT/'+detector_name+'/DET/image/mask', layouts['mask'], fillvalue=np.nan)
                    outf.create_virtual_dataset('INSTRUMENT/'+detector_name+'/DET/image/gain', layouts['gain'], fillvalue=np.nan)                
                outf.close()
        def generate_module_wise_vds_files(module_id,results,module_ids,n_good_frames,local_good_frame_ids_by_part,good_frame_slices,**args):
            result_ids = np.nonzero(module_ids==module_id)[0]
            #log.info('module_id = {}'.format(module_id))
            layouts = {
                'data':self.h5_VirtualLayout(shape=(n_good_frames,) + data_shape[1:], dtype = data_dtype),
                'mask':self.h5_VirtualLayout(shape=(n_good_frames,) + data_shape[1:], dtype = mask_dtype),
                'gain':self.h5_VirtualLayout(shape=(n_good_frames,) + data_shape[1:], dtype = gain_dtype)
            }
            #log.info('data layout shape = {}'.format(layouts['data'].shape))
            cell_ids = np.zeros((n_good_frames,),dtype= cell_dtype)
            pulse_ids = np.zeros((n_good_frames,),dtype= pulse_dtype)
            train_ids = np.zeros((n_good_frames,),dtype= train_dtype)
            for r_id in result_ids:
                result = results[r_id]
                part_id = result['part_id']
                good_frame_slice = good_frame_slices[part_id]
                local_good_frame_ids = local_good_frame_ids_by_part[part_id]
                populate_vds(result,layouts,local_good_frame_ids,good_frame_slice,index_module = False)
                cell_ids[good_frame_slice] = result['cells']
                pulse_ids[good_frame_slice] = result['pulses']
                train_ids[good_frame_slice] = result['trains']
            
            vds_file_name = self.get_path('vds_module',path_modifiers={'run':run,'data_mode':data_mode,'module':module_id})
            #log.info('vds_file name = {}'.format(vds_file_name))
            save_vds_file(layouts,cell_ids,pulse_ids,train_ids,n_good_frames,vds_file_name)
            #log.info('done')            
            return None
        
        def populate_vds(result,layouts,local_good_frame_ids,good_frame_slice,index_module=True):
            module = result['module']
            part_id = result['part_id']
            frame_slice = frame_slices[part_id]
            vsource_data = result['v_source']
            layout_data = layouts['data']
            #log.info('slice in populate_vds ={}'.format(good_frame_slice))
            #log.info('max_local frame_id ={}'.format(local_good_frame_ids.max()))
            #log.info('layout  shape {}'.format(layout_data.shape))
            #log.info('vsource shape {}'.format(vsource_data.shape))             
            if index_module:
                layout_data[module,good_frame_slice] = vsource_data[local_good_frame_ids]
            else:
                #log.info('vsource_shape = {}'.format(vsource_data.shape))
                #log.info('vsource_shape = {}'.format(vsource_data.shape))
                #log.info('layout_shape = {}'.format(layout_data[good_frame_slice].shape))
                #layout_data[good_frame_slice] = vsource_data[combined_good_frames_part]
                layout_data[good_frame_slice] = vsource_data[local_good_frame_ids]
            if not from_raw_data:
                layout_mask = layouts['mask']
                vsource_mask = result['v_source_mask']                            
                layout_gain = layouts['gain']
                vsource_gain = result['v_source_gain']
                #log.info('layout gain shape {}'.format(layout_gain.shape))
                #log.info('vsource gain shape {}'.format(vsource_gain.shape))
                if index_module:
                    layout_mask[module,good_frame_slice] = vsource_mask[local_good_frame_ids]
                    layout_gain[module,good_frame_slice] = vsource_gain[local_good_frame_ids]
                else:
                    #log.info('len frame_ids={}'.format(len(frame_ids)))
                    layout_mask[good_frame_slice] = vsource_mask[local_good_frame_ids]
                    layout_gain[good_frame_slice] = vsource_gain[local_good_frame_ids]

        
        n_files = len(flist)
        stime = time.time()
        
        #log.info('creating_vds in multiprocessing mode')
        results = comm_module.request_mp_evaluation(vds_worker,argArrays=[flist,m_ids],split_together = True,assemble_outputs = False,n_processes = n_processes)
        #log.info('len(m_ids) = {}'.format(len(m_ids)))
        results = tuple( elem for result in results.values() for elem in result)

        # Calculate frames that are not mask for all modules and use them to generate the vds File
        combined_frame_mask = np.full(n_frames,True)
        for r in results:
            combined_frame_mask[frame_slices[r['part_id']]] &= r['local_frame_mask']
        local_good_frame_ids_by_part = []
        for s in frame_slices:
            local_n_frames = s.stop-s.start
            local_frame_ids = np.arange(local_n_frames)
            local_good_frame_ids_by_part.append(local_frame_ids[combined_frame_mask[s]])
        
        # Calculate slices in combined_good_frames that correspond to those in frame_slices
        n_good_frames_per_slice = [np.sum(combined_frame_mask[s]) for s in frame_slices]
        #log.info('n_good_frames_per_slice ={}'.format(n_good_frames_per_slice))
        temp = [np.sum(n_good_frames_per_slice[:i]) for i in range(len(frame_slices)+1)]
        #log.info('temp ={}'.format(temp))
        good_frame_slices = [slice(int(temp[i]),int(temp[i+1])) for i in range(len(frame_slices))]            
        #log.info('good_frame_slices ={}'.format(good_frame_slices))
        n_good_frames = np.sum(n_good_frames_per_slice)
        log.info('From {} inital frames {} remain after masking, that is {}%'.format(n_frames,n_good_frames,n_good_frames/n_frames*100))
        
        if create_modulewise_vds:
            #layouts_per_module = comm_module.request_mp_evaluation(generate_module_wise_vds_files,argArrays=[np.array([0,1,2,3,4,5,6,7,8,9,10,11,12,13,14,15])],const_args = [results,m_ids,n_good_frames,local_good_frame_ids_by_part,good_frame_slices] ,assemble_outputs = False,n_processes = n_processes)
            layouts_per_module = comm_module.request_mp_evaluation(generate_module_wise_vds_files,argArrays=[modules],const_args = [results,m_ids,n_good_frames,local_good_frame_ids_by_part,good_frame_slices] ,assemble_outputs = False,n_processes = n_processes)
        if create_complete_vds:
            raise NotImplementedError('omplete vds not implemented yet.')


    ## misc saving routines ##
    def save_hit_specifier_vds(self,name,data,**kwargs):
        runs = opt['runs'] #iterable
        modules = opt['modules']
        info = self.load_info({'runs':runs})
        n_frames=tuple(info[str(run)]['n_frames'] for run in runs)
        n_frames_total = np.sum(n_frames)
        l_hit_specifiers = tuple(self.h5_VirtualLayout(shape=(n_frames_total,3), dtype=float) for module in modules)

        runs_array = np.concatenate(
            tuple(
                np.concatenate(( np.full((n_frames[id],),run)[:,None] , np.arange(n_frames[id])[:,None]),axis=1) for id,run in enumerate(runs)),axis = 0)
        
        current_frame = 0
        for run in runs :
            filename = super().get_path('hit_specifier',path_modifiers={'run':'%04d'%run})
            length = info[str(run)]['n_frames']            
            vsrc_hit_specifiers = tuple( self.h5_VirtualSource(filename, 'litpixels_{}'.format('%02d'%module), shape=(length,)) for module in modules)
            c=current_frame
            for m_id in range(len(modules)):
                l_hit_specifiers[m_id][c:c+length] = vsrc_hit_specifiers[m_id]
            current_frame += length
            
        value_dict={
            'module_%02d'%module:l_hit_specifiers[m_id] for m_id,module in enumerate(modules)
        }
        value_dict['runs']=runns_array

        super().save('vds_hit_specifier',value_dict)
        
    def save_hits_vds(self,name,data,**kwargs):
        runs = opt['runs'] #iterable
        n_hits=[]
        for run in runs:
            with  self.load_direct('hits',path_modifiers={'run':run},as_h5_object = True) as h5_file:
                n_hits.append(h5_file['hits/assembled'].shape[0])

        n_hits_total = np.sum(n_hits)
        l_hit_specifiers = tuple(self.h5_VirtualLayout(shape=(n_hits_total,1296,1138), dtype=float) for module in modules)

        runs_array = np.concatenate(tuple( np.concatenate((np.full((n_frames[id],),run)[:,None],np.arange(n_frames[id])[:,None]),axis=1) for id,run in enumerate(runs)),axis = 0)
        current_frame = 0
        for run in runs :
            filename = super().get_path('hit_specifier',path_modifiers={'run':'%04d'%run})
            length = info[str(run)]['n_frames']            
            vsrc_hit_specifiers = tuple( self.h5_VirtualSource(filename, 'litpixels_{}'.format('%02d'%module), shape=(length,)) for module in modules)
            c = current_frame
            for m_id in range(len(modules)):
                l_hit_specifiers[m_id][c:c+length] = vsrc_hit_specifiers[m_id]
            current_frame += length
            
        value_dict={
            'module_%02d'%module:l_hit_specifiers[m_id] for m_id,module in enumerate(modules)
        }
        value_dict['runs']=runns_array

        super().save('vds_hit_specifier',value_dict)
